
"""Jaime Sendra Berenguer-2020.
MLearner Machine Learning Library Extensions
Author:Jaime Sendra Berenguer<www.linkedin.com/in/jaisenbe>
License: MIT
"""

import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin


class FeatureSelector(BaseEstimator, TransformerMixin):

    """This transformer select features.

    Attributes
    ----------
    columns: list of columns to transformer [n_columns]

    Examples
    --------
    For usage examples, please see
    https://jaisenbe58r.github.io/MLearner/user_guide/preprocessing/FeatureSelector/

    """
    def __init__(self, columns=None, random_state=99):
        """Init replace missing values."""
        if columns is not None:
            if isinstance(columns, list) or isinstance(columns, tuple):
                self.columns = columns
            else:
                raise NameError("Invalid type {}".format(type(columns)))
        else:
            self.columns = columns

        self.random_state = random_state

    def fit(self, X, y=None, **fit_params):
        """Gets the columns to make a replace missing values.

        Parameters
        ----------
        X : {Dataframe}, shape = [n_samples, n_features]
            Dataframe, where n_samples is the number of samples and
            n_features is the number of features.

        Returns
        --------
        self

        """
        if not isinstance(X, pd.core.frame.DataFrame):
            raise NameError("Invalid type {}".format(type(X)))

        if self.columns is None:
            self.columns = X.columns

        _lista = [i for i in self.columns if i not in X.columns.tolist()]
        if len(_lista) > 0:
            raise NameError("The columns {} no exist in Dataframe".format(_lista))
        self._fitted = True
        return self

    def transform(self, X):
        """this transformer handles missing values.

        Parameters
        ----------
        X : {Dataframe}, shape = [n_samples, n_features]
            Dataframe of samples, where n_samples is the number of samples and
            n_features is the number of features.

        Returns
        -------
        X : {Dataframe}, shape = [n_samples, n_features]
            A copy of the input Dataframe with the columns replaced.

        """
        if not hasattr(self, "_fitted"):
            raise AttributeError("FeatureSelector has not been fitted, yet.")

        if not isinstance(X, pd.core.frame.DataFrame):
            raise NameError("Invalid type {}".format(type(X)))

        return X[self.columns]


class DataFrameSelector(BaseEstimator, TransformerMixin):
    def __init__(self, attribute_names):
        self.attribute_names = attribute_names

    def fit(self, X, y=None):
        return self

    def transform(self, X):
        return X[self.attribute_names]


class CopyFeatures(BaseEstimator, TransformerMixin):
    def __init__(self, columns=None, prefix=""):
        """Ini copy features."""
        if columns is not None:
            if isinstance(columns, list) or isinstance(columns, tuple):
                self.columns = columns
            else:
                raise TypeError("Invalid type {}".format(type(columns)))
        else:
            self.columns = columns

        self.name = prefix

    def fit(self, X, y=None):
        if self.columns is None:
            self.columns = X.select_dtypes(exclude=["object"]).columns

        if isinstance(X, pd.core.frame.DataFrame):
            try:
                _test = X[self.columns].astype(np.float32)
                del(_test)
            except ValueError:
                raise NameError("Null or categorical variables are not allowed: {}".format(X.dtypes))
        else:
            raise NameError("Invalid type {}".format(type(X)))
        self._fitted = True

        return self

    def transform(self, X):
        if not hasattr(self, "_fitted"):
            raise AttributeError("CopyFeatures has not been fitted, yet.")

        if not isinstance(X, pd.core.frame.DataFrame):
            raise NameError("Invalid type {}".format(type(X)))

        self.X_transform = X.copy()
        for i in self.columns:
            name_col = self.name + "_" + str(i)
            self.X_transform[name_col] = X[i].values

        return self.X_transform


class DropFeatures(BaseEstimator, TransformerMixin):

    """This transformer drop features.

    Attributes
    ----------
    columns: list of columns to transformer [n_columns]

    Examples
    --------
    For usage examples, please see
    https://jaisenbe58r.github.io/MLearner/user_guide/preprocessing/DropFeatures/

    """
    def __init__(self, columns_drop=None, random_state=99):
        """Init replace missing values."""
        if columns_drop is not None:
            if isinstance(columns_drop, list) or isinstance(columns_drop, tuple):
                self.columns_drop = columns_drop
            else:
                raise NameError("Invalid type {}".format(type(columns_drop)))
        else:
            self.columns_drop = columns_drop

        self.random_state = random_state

    def fit(self, X, y=None, **fit_params):
        """Gets the columns to make a replace missing values.

        Parameters
        ----------
        X : {Dataframe}, shape = [n_samples, n_features]
            Dataframe, where n_samples is the number of samples and
            n_features is the number of features.

        Returns
        --------
        self

        """
        if not isinstance(X, pd.core.frame.DataFrame):
            raise NameError("Invalid type {}".format(type(X)))

        if self.columns_drop is None:
            self.columns_drop = X.columns_drop

        _lista = [i for i in self.columns_drop if i not in X.columns.tolist()]
        if len(_lista) > 0:
            raise NameError("The columns {} no exist in Dataframe".format(_lista))
        self._fitted = True
        return self

    def transform(self, X):
        """this transformer handles missing values.

        Parameters
        ----------
        X : {Dataframe}, shape = [n_samples, n_features]
            Dataframe of samples, where n_samples is the number of samples and
            n_features is the number of features.

        Returns
        -------
        X : {Dataframe}, shape = [n_samples, n_features]
            A copy of the input Dataframe with the columns replaced.

        """
        if not hasattr(self, "_fitted"):
            raise AttributeError("DropFeatures has not been fitted, yet.")

        if not isinstance(X, pd.core.frame.DataFrame):
            raise NameError("Invalid type {}".format(type(X)))

        return X.drop(self.columns_drop, axis=1)



"""
Jaime Sendra Berenguer-2020.
MLearner Machine Learning Library Extensions
Author:Jaime Sendra Berenguer<www.linkedin.com/in/jaisenbe>
License: MIT
"""

from .ml_evaluation import EvaluationModels


__all__ = ["EvaluationModels"]



"""
Jaime Sendra Berenguer-2020.
MLearner Machine Learning Library Extensions
Author:Jaime Sendra Berenguer<www.linkedin.com/in/jaisenbe>
License: MIT
"""

from .ml_neural import Neural, Neural_sklearn

__all__ = ["Neural", "Neural_sklearn"]



"""
Jaime Sendra Berenguer-2020.
MLearner Machine Learning Library Extensions
Author:Jaime Sendra Berenguer<www.linkedin.com/in/jaisenbe>
License: MIT
"""

import pandas as pd
import re
import emoji
from sklearn.feature_extraction.text import CountVectorizer

"""
Regex Helpers

Major RE functions

re.findall - Module is used to search for “all” occurrences that match a given pattern.
re.sub - Substitute the matched RE patter with given text
re.match - The match function is used to match the RE pattern to string with optional flags
re.search - This method takes a regular expression pattern and a string and searches for that pattern with the string.
We will be mostly using re.findall to detect patterns.

Based in https://www.kaggle.com/raenish/cheatsheet-text-helper-functions
"""


def find_url(string):
    """
    Search URL in text.

    Parameters
    ----------
        string: str
            Text selected to apply transformation

    Examples:
    ---------
    ```python
    sentence="I love spending time at https://www.kaggle.com/"
    find_url(sentence)
    ```
    """
    text = re.findall('http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*\(\),]|(?:%[0-9a-fA-F][0-9a-fA-F]))+', string)
    return "".join(text)  # converting return value from list to string.


def find_emoji(text):
    """
    Find and convert emoji to text.

    Parameters
    ----------
        text: str
            Text selected to apply transformation

    Examples:
    ---------
    ```python
    sentence="I love () very much ())"
    find_emoji(sentence)
        >>> ['soccer_ball', 'beaming_face_with_smiling_eyes']
    ```
    """
    emo_text = emoji.demojize(text)
    return re.findall(r'\:(.*?)\:', emo_text)


# def remove_emoji(text):
#     """
#     Remove Emoji from text.

#     Parameters
#     ----------
#         text: str
#             Text selected to apply transformation

#     Examples:
#     ---------
#     ```python
#     sentence="Its all about \U0001F600 face"
#     print(sentence)
#     remove_emoji(sentence)
#         >>> Its all about 😀 face
#         >>> 'Its all about  face'
#     ```
#     """
#     emoji_pattern = re.compile("["
#                                 u"\U0001F600-\U0001F64F"  # emoticons
#                                 u"\U0001F300-\U0001F5FF"  # symbols & pictographs
#                                 u"\U0001F680-\U0001F6FF"  # transport & map symbols
#                                 u"\U0001F1E0-\U0001F1FF"  # flags (iOS)
#                                 u"\U00002702-\U000027B0"
#                                 u"\U000024C2-\U0001F251"
#                                 "]+", flags=re.UNICODE)
#     return emoji_pattern.sub(r'', text)


def find_email(text):
    """
    Extract email from text.

    Parameters
    ----------
        text: str
            Text selected to apply transformation

    Examples:
    ---------
    ```python
    sentence="My gmail is abc99@gmail.com"
    find_email(sentence)
        >>> 'abc99@gmail.com'
    ```
    """
    line = re.findall(r'[\w\.-]+@[\w\.-]+', str(text))
    return ",".join(line)


def find_hash(text):
    """
    This value is especially to denote trends in twitter.

    Parameters
    ----------
        text: str
            Text selected to apply transformation

    Examples:
    ---------
    ```python
    sentence="#Corona is trending now in the world"
    find_hash(sentence)
        >>> 'Corona'
    ```
    """
    line = re.findall(r'(?<=#)\w+', text)
    return " ".join(line)


def find_at(text):
    """
    @ - Used to mention someone in tweets

    Parameters
    ----------
        text: str
            Text selected to apply transformation

    Examples:
    ---------
    ```python
    sentence="@David,can you help me out"
    find_at(sentence)
        >>> 'David'
    ```
    """
    line = re.findall(r'(?<=@)\w+', text)
    return " ".join(line)


def find_number(text):
    """
    Pick only number from sentence

    Parameters
    ----------
        text: str
            Text selected to apply transformation

    Examples:
    ---------
    ```python
    sentence="2833047 people are affected by corona now"
    find_number(sentence)
        >>> '2833047'
    ```
    """
    line = re.findall(r'[0-9]+', text)
    return " ".join(line)


def find_phone_number(text):
    """
    Spain Mobile numbers have ten digit. I will write that pattern below.

    Parameters
    ----------
        text: str
            Text selected to apply transformation

    Examples:
    ---------
    ```python
    find_phone_number("698887776 is a phone number of Mark from 210,North Avenue")
        >>> '698887776'
    ```
    """
    line = re.findall(r"(\+34|0034|34)?[ -]*(6|7)[ -]*([0-9][ -]*){8}", text)
    return line


def find_year(text):
    """
    Extract year from 1940 till 2040.

    Parameters
    ----------
        text: str
            Text selected to apply transformation

    Examples:
    ---------
    ```python
    sentence="India got independence on 1947."
    find_year(sentence)
        >>> ['1947']
    ```
    """
    line = re.findall(r"\b(19[40][0-9]|20[0-1][0-9]|2040)\b", text)
    return line


def find_nonalp(text):
    """
    Extract Non Alphanumeric characters.

    Parameters
    ----------
        text: str
            Text selected to apply transformation

    Examples:
    ---------
    ```python
    sentence="Twitter has lots of @ and # in posts.(general tweet)"
    find_nonalp(sentence)
        >>> ['@', '#', '.', '(', ')']
    ```
    """
    line = re.findall("[^A-Za-z0-9 ]", text)
    return line


def find_punct(text):
    """
    Retrieve punctuations from sentence.

    Parameters
    ----------
        text: str
            Text selected to apply transformation

    Examples:
    ---------
    ```python
    example="Corona virus have kiled #24506 confirmed cases now.#Corona is un(tolerable)"
    print(find_punct(example))
        >>> ['#', '.', '#', '(', ')']
    ```
    """
    line = re.findall(r'[!"\$%&\'()*+,\-.\/:;=#@?\[\\\]^_`{|}~]*', text)
    string = "".join(line)
    return list(string)


def ngrams_top(corpus, ngram_range, n=None, idiom='english'):
    """
    List the top n words in a vocabulary according to occurrence in a text corpus.

    Examples:
    ---------
    ```python
    ngrams_top(df['text'],(1,1),n=10)
        >>> 	text	count
        >>> 0	just	2278
        >>> 1	day     2115
        >>> 2	good	1578
        >>> 3	like	1353
        >>> 4	http	1247
        >>> 5	work	1150
        >>> 6	today	1147
        >>> 7	love	1145
        >>> 8	going	1103
        >>> 9	got	    1085
    ngrams_top(df['text'],(2,2),n=10)
        >>> 	text	        count
        >>> 0	mother day	    358
        >>> 1	twitpic com	    334
        >>> 2	http twitpic    332
        >>> 3	mothers day	    279
        >>> 4	happy mother    275
        >>> 5	just got	    219
        >>> 6	happy mother    199
        >>> 7	http bit	    180
        >>> 8	bit ly	        180
        >>> 9	good morning    176
    ```
    """
    vec = CountVectorizer(stop_words=idiom, ngram_range=ngram_range).fit(corpus)
    bag_of_words = vec.transform(corpus)
    sum_words = bag_of_words.sum(axis=0)
    words_freq = [(word, sum_words[0, idx]) for word, idx in vec.vocabulary_.items()]
    words_freq = sorted(words_freq, key=lambda x: x[1], reverse=True)
    total_list = words_freq[:n]
    df = pd.DataFrame(total_list, columns=['text', 'count'])
    return df


def unique_char(sentence):
    """
    Retrieve punctuations from sentence.
    If you want to change match repetitive characters to n numbers,
    chage the return line in the rep function to grp[0:n].

    Parameters
    ----------
        sentence: str
            Text selected to apply transformation

    Examples:
    ---------
    ```python
    sentence="heyyy this is loong textttt sooon"
    unique_char(sentence)
        >>> 'hey this is long text son'
    ```
    """
    def rep(text):
        grp = text.group(0)
        if len(grp) > 1:
            return grp[0:1]  # can change the value here on repetition

    convert = re.sub(r'(\w)\1+', rep, sentence)
    return convert


def find_coin(text, symbol="$"):
    """
    Find prices in text

    Parameters
    ----------
        text: str
            Text selected to apply transformation

        symbol: str
            Coin symbol

    Examples:
    ---------
    ```python
    sentence="this shirt costs $20.56"
    find_dollar(sentence)
        >>> '$20.56'
    ```
    """
    if symbol == "$":
        line = re.findall(r'\$\d+(?:\.\d+)?', text)
    elif symbol == "€":
        line = re.findall(r'\€\d+(?:\.\d+)?', text)
    else:
        raise NameError(f"Coin symbol {symbol} not implemented")
    return " ".join(line)


def num_great(text):
    """
    Number greater than 930

    Parameters
    ----------
        text: str
            Text selected to apply transformation

    Examples:
    ---------
    ```python
    sentence="It is expected to be more than 935 corona death
        and 29974 observation cases across 29 states in india"
    num_great(sentence)
        >>> '935 29974'
    ```
    """
    line = re.findall(r'9[3-9][0-9]|[1-9]\d{3,}', text)
    return " ".join(line)


def num_less(text):
    """
    Number less than 930.

    Parameters
    ----------
        text: str
            Text selected to apply transformation

    Examples:
    ---------
    ```python
    sentence="There are some countries where less than 920 cases
        exist with 1100 observations"
    num_less(sentence)
        >>> '920'
    ```
    """
    only_num = []
    for i in text.split():
        line = re.findall(r'^(9[0-2][0-0]|[1-8][0-9][0-9]|[1-9][0-9]|[0-9])$', i)
        only_num.append(line)
        all_num = [",".join(x) for x in only_num if x != []]
    return " ".join(all_num)


def find_dates(text):
    """
    Find Dates. mm-dd-yyyy format

    Parameters
    ----------
        text: str
            Text selected to apply transformation

    Examples:
    ---------
    ```python
    sentence="Todays date is 04/28/2020 for format mm/dd/yyyy, not 28/04/2020"
    find_dates(sentence)
        >>> [('04', '28', '2020')]
    ```
    """
    line = re.findall(r'\b(1[0-2]|0[1-9])/(3[01]|[12][0-9]|0[1-9])/([0-9]{4})\b', text)
    return line


def only_words(text):
    """
    Only Words - Discard Numbers.

    Parameters
    ----------
        text: str
            Text selected to apply transformation

    Examples:
    ---------
    ```python
    sentence="the world population has grown from 1650 million to 6000 million"
    only_numbers(sentence)
        >>> '1650 6000'
    ```
    """
    line = re.findall(r'\b[^\d\W]+\b', text)
    return " ".join(line)


def boundary(text):
    """
    Extracting word with boundary

    Parameters
    ----------
        text: str
            Text selected to apply transformation

    Examples:
    ---------
    ```python
    sentence="Most tweets are neutral in twitter"
    boundary(sentence)
        >>> 'neutral'
    ```
    """
    line = re.findall(r'\bneutral\b', text)
    return " ".join(line)


def search_string(text, key):
    """
    Is the key word present in the sentence?

    Parameters
    ----------
        text: str
            Text selected to apply transformation

        key: str
            Word to search within the phrase

    Examples:
    ---------
    ```python
    sentence="Happy Mothers day to all Moms"
    search_string(sentence,'day')
        >>> True
    ```
    """
    return bool(re.search(r''+key+'', text))


def pick_only_key_sentence(text, keyword):
    """
    If we want to get all sentence with particular keyword.
    We can use below function.

    Parameters
    ----------
        text: str
            Text selected to apply transformation.

        keyword: str
            Word to search within the phrase.

    Examples:
    ---------
    ```python
    sentence="People are fighting with covid these days.
        Economy has fallen down.How will we survice covid"
    pick_only_key_sentence(sentence,'covid')
        >>> ['People are fighting with covid these days',
            'How will we survice covid']
    ```
    """
    line = re.findall(r'([^.]*'+keyword+'[^.]*)', text)
    return line


def pick_unique_sentence(text):
    """
    Most webscrapped data contains duplicated sentence.
    This function could retrieve unique ones.

    Parameters
    ----------
        text: str
            Text selected to apply transformation.

        keyword: str
            Word to search within the phrase.

    Examples:
    ---------
    ```python
    sentence="I thank doctors\nDoctors are working very hard
        in this pandemic situation\nI thank doctors"
    pick_unique_sentence(sentence)
        >>> ['Doctors are working very hard in this
            pandemic situation', 'I thank doctors']
    ```
    """
    line = re.findall(r'(?sm)(^[^\r\n]+$)(?!.*^\1$)', text)
    return line


def find_capital(text):
    """
    Extract words starting with capital letter.
    Some words like names,place or universal object
    are usually mentioned in a text starting with CAPS.

    Parameters
    ----------
        text: str
            Text selected to apply transformation.

    Examples:
    ---------
    ```python
    sentence="World is affected by corona crisis.
        No one other than God can save us from it"
    find_capital(sentence)
        >>> ['World', 'No', 'God']
    ```
    """
    line = re.findall(r'\b[A-Z]\w+', text)
    return line


def remove_tag(string):
    """
    Most of web scrapped data contains html tags.
    It can be removed from below re script

    Parameters
    ----------
        text: str
            Text selected to apply transformation.

    Examples:
    ---------
    ```python
    sentence="Markdown sentences can use <br> for breaks and <i></i> for italics"
    remove_tag(sentence)
        >>> 'Markdown sentences can use  for breaks and  for italics'
    ```
    """
    text = re.sub('<.*?>', '', string)
    return text


def mac_add(string):
    """
    Extract Mac address from text.
    https://stackoverflow.com/questions/26891833/python-regex-extract-mac-addresses-from-string/2689237

    Parameters
    ----------
        string: str
            Text selected to apply transformation.

    Examples:
    ---------
    ```python
    sentence="MAC ADDRESSES of this laptop - 00:24:17:b1:cc:cc .
        Other details will be mentioned"
    mac_add(sentence)
        >>> ['00:24:17:b1:cc:cc']
    ```
    """
    text = re.findall('(?:[0-9a-fA-F]:?){12}', string)
    return text


def ip_add(string):
    """
    Extract IP address from text.

    Parameters
    ----------
        string: str
            Text selected to apply transformation.

    Examples:
    ---------
    ```python
    sentence="An example of ip address is 125.16.100.1"
    ip_add(sentence)
        >>> ['125.16.100.1']
    ```
    """
    text = re.findall('\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}', string)
    return text


def subword(string, sub):
    """
    Extract number of subwords from sentences and words.

    Parameters
    ----------
        string: str
            Text selected to apply transformation.

        sub: str
            subwords from sentences

    Examples:
    ---------
    ```python
    sentence = 'Fundamentalism and constructivism are important skills'
    subword(sentence,'ism') # change subword and try for others
        >>> 2
    ```
    """
    text = re.findall(sub, string)
    return len(text)


def lat_lon(string, display=False):
    """
    valid latitude & longitude

    Parameters
    ----------
        string: str
            Text selected to apply transformation.

    Examples:
    ---------
    ```python
    lat_lon('28.6466772,76.8130649')
    lat_lon('2324.3244,3423.432423')
        >>> [28.6466772,76.8130649] is valid latitude & longitude
        >>> [2324.3244,3423.432423] is not a valid latitude & longitude
    ```
    """
    text = re.findall(r'^[-+]?([1-8]?\d(\.\d+)?|90(\.0+)?),\s*[-+]?(180(\.0+)?|((1[0-7]\d)|([1-9]?\d))(\.\d+)?)$', string)
    if text != []:
        if display:
            print("[{}] is valid latitude & longitude".format(string))
        return True
    else:
        if display:
            print("[{}] is not a valid latitude & longitude".format(string))
        return False


def pos_look_ahead(string, A, B):
    """
    Positive look ahead will succeed if passed non-consuming expression
    does match against the forthcoming input.
    The syntax is A(?=B) where A is actual expression and B
    is non-consuming expression.

    Parameters
    ----------
        string: str
            Text selected to apply transformation.

        A, B: str
            A is actual expression and B is non-consuming expression.

    Examples:
    ---------
    ```python
    pos_look_ahead("I love kaggle. I love DL","love","DL")
        >>> position:(17, 21) Matched word:love
    ```
    """
    pattern = re.compile(''+A+'(?=\s'+B+')')
    match = pattern.search(string)
    print("position:{} Matched word:{}".format(match.span(), match.group()))


def neg_look_ahead(string, A, B):
    """
    Negative look ahead will succeed if passed non-consuming expression
    does not match against the forthcoming input.
    The syntax is A(?!B) where A is actual expression and B
    is non-consuming expression.

    Parameters
    ----------
        string: str
            Text selected to apply transformation.

        A, B: str
            A is actual expression and B is non-consuming expression.

    Examples:
    ---------
    ```python
    neg_look_ahead("I love kaggle. I love DL","love","DL")
        >>> position:(2, 6) Matched word:love
    ```
    """
    pattern = re.compile(''+A+'(?!\s'+B+')')
    match = pattern.search(string)
    print("position:{} Matched word:{}".format(match.span(), match.group()))


def pos_look_behind(string, A, B):
    """
    Positive look behind will succeed if passed non-consuming expression
    does match against the forthcoming input.
    The syntax is A(?<=B) where A is actual expression and B
    is non-consuming expression.

    Parameters
    ----------
        string: str
            Text selected to apply transformation.

        A, B: str
            A is actual expression and B is non-consuming expression.

    Examples:
    ---------
    ```python
    pos_look_behind("i love nlp.everyone likes nlp","love","nlp")
    # the word "nlp" that do come after "love"
        >>> position:(7, 10) Matched word: nlp
    ```
    """
    pattern = re.compile("(?<="+A+"\s)"+B+"")
    match = pattern.search(string)
    print("position:{} Matched word: {}".format(match.span(), match.group()))


def neg_look_behind(string, A, B):
    """
    Positive look behind will succeed if passed non-consuming expression
    does not match against the forthcoming input.
    The syntax is "A(?<!=B)" where "A"is actual expression and "B"
    is non-consuming expression.

    Parameters
    ----------
        string: str
            Text selected to apply transformation.

        A, B: str
            A is actual expression and B is non-consuming expression.

    Examples:
    ---------
    ```python
    neg_look_behind("i love nlp.everyone likes nlp","love","nlp")
    # the word "nlp" that doesnt come after "love"
        >>> position:(26, 29) Matched word: nlp
    ```
    """
    pattern = re.compile("(?<!"+A+"\s)"+B+"")
    match = pattern.search(string)
    print("position:{} Matched word: {}".format(match.span(), match.group()))


def find_domain(string):
    """
    Search domains in the text.

    Parameters
    ----------
        string: str
            Text selected to apply transformation.

        A, B: str
            A is actual expression and B is non-consuming expression.

    Examples:
    ---------
    ```python
    sentence="WHO provides valid information about covid in their site who.int.
        UNICEF supports disadvantageous childrens. know more in unicef.org"
    find_domain(sentence)
        >>> ['who.int', 'unicef.org']
    ```
    """
    text = re.findall(r'\b(\w+[.]\w+)', string)
    return text



"""
Jaime Sendra Berenguer-2020.
MLearner Machine Learning Library Extensions
Author:Jaime Sendra Berenguer<www.linkedin.com/in/jaisenbe>
License: MIT
"""

from .params_manager import ParamsManager
from .keras import keras_checkpoint, MyCustomCallback, EarlyStoppingAtMinLoss, \
                    LearningRateScheduler

__all__ = ["ParamsManager", "keras_checkpoint", "MyCustomCallback",
            "EarlyStoppingAtMinLoss", "LearningRateScheduler"]




"""Jaime Sendra Berenguer-2020.
MLearner Machine Learning Library Extensions
Author:Jaime Sendra Berenguer<www.linkedin.com/in/jaisenbe>
License: MIT
"""

import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import time
import pickle
import os

from sklearn.model_selection import StratifiedKFold, cross_val_score, train_test_split
from sklearn.model_selection import GridSearchCV, RandomizedSearchCV
from sklearn.naive_bayes import GaussianNB
from sklearn.neighbors import KNeighborsClassifier
from sklearn.svm import SVC

from sklearn.ensemble import RandomForestClassifier
from sklearn.metrics import accuracy_score
from sklearn.inspection import permutation_importance
from sklearn.feature_selection import SelectFromModel
from sklearn.pipeline import Pipeline

from lightgbm import LGBMClassifier
from xgboost import XGBClassifier

from mlearner.training import Training
from mlearner.models import modelCatBoost, modelLightBoost, modelXGBoost
from mlearner.utils import ParamsManager
from sklearn.base import BaseEstimator, TransformerMixin

import warnings
warnings.filterwarnings("ignore")

param_file = "mlearner/classifier/config/models.json"


class TrainingUtilities(object):
    def __init__(self, random_state=99):
        """
        Inicializacion de la clase de Preprocesado de un dataframe.
        """
        self.random_state = random_state
        self.manager_models = ParamsManager(param_file, key_read="Models")

    def Seleccion_rasgos(self, model, X, Y, n, display=True):
        """
        Funcion que nos retorna las categorias que mas influyen en el modelo.

        Inputs:
            - X, Y: dataset.
            - n: numero de categorias que queremos que formen parte de variables del modelo de RL.
            - display: Mostrar los resultados por pantalla

        Outputs:
            - z: Resumen de la inferencia de cada categoria en el modelo de RL.
            - valid: Categorias seleccionadas para formar parte del modelo de RL.
        """
        from sklearn.feature_selection import RFE

        categorias = list(X.columns)

        rfe = RFE(model, n)
        rfe = rfe.fit(X, Y.values.ravel())

        z = zip(X, rfe.support_, rfe.ranking_)
        categorias = list(X.columns)
        valid = []

        for i in range(len(rfe.support_)):

            if rfe.support_[i]:
                valid.append(categorias[i])

        if display:
            print("Resumen de las iteraciones para todas las categorias: \n")
            print(list(z))
            print("\n")
            print("Categorias seleccionadas: \n")
            print(valid)

        return z, valid

    def optimizacion_seleccion_rasgos(self, model, X, Y, n_splits, display=True):
        """
        Seleccion del numero de categorias a tener en cuenta
        """
        scores = []
        nums = []
        for i in range(len(X.columns)):

            _, valid = self.Seleccion_rasgos(model, X, Y, i+1)
            sc = self.validacion_cruzada(model, X[valid], Y, n_splits, shuffle=True)
            scores.append(sc)
            nums.append(i+1)

        d = {'n': nums, 'score': scores}
        df = pd.DataFrame(data=d)
        if display:
            plt.plot(df.n, df.score, marker="o", linestyle="--", color="r")
            plt.xlabel("numero categorias")
            plt.ylabel("Score")
            plt.grid(True)
            plt.title("Score frente al numero de categorias")

        return df

    def validacion_cruzada(self, model, X, Y, n_splits, shuffle=True, scoring="accuracy"):
        """
        Validacion cruzada del dataset introducido como input.

        Inputs:
            - cv = Numero de iteraciones.

        Outputs:
            - score: Media de los acc de todas las iteraciones.
        """
        from sklearn.model_selection import cross_val_score
        from sklearn.model_selection import KFold

        cv = KFold(n_splits=n_splits, shuffle=True, random_state=self.random_state)
        scores = cross_val_score(model, X, Y, scoring=scoring, cv=cv)

        return scores.mean()

    def opt_RandomForest_Classifier(self, X, Y, nmin=1, nmax=10000, num=4, display=True):
        """
        Seleccion del numero de categorias a tener en cuenta para RF
        """
        X_train, X_test, y_train, y_test = train_test_split(X, Y, test_size=0.2)
        rango = np.geomspace(nmin, nmax, num=num, endpoint=True)
        scores = []
        scores_val = []
        nums = []
        forestal = []

        for i in list(rango):

            forest = RandomForestClassifier(X_train, y_train, int(i))
            scores.append(forest.oob_score_)
            nums.append(int(i))
            forestal.append(forest)
            scores_val.append(forest.score(X_test, y_test))

        d = {'n': nums, 'score': scores, 'score_val': scores_val}
        df = pd.DataFrame(data=d)

        if display:

            plt.semilogx(df.n, df.score, marker="o", linestyle="--", color="r")
            plt.semilogx(df.n, df.score_val, marker="o", linestyle="--", color="b")
            plt.xlabel("numero categorias")
            plt.ylabel("Score")
            plt.grid(True)
            plt.title("Score frente al numero de categorias")

        return df, forestal


class PipelineClasificators(Training):

    def __init__(self, random_state=99):
        """
        Inicializacion de la clase de modelos
        """
        self.random_state = random_state
        self.manager_models = ParamsManager(param_file, key_read="Models")
        self.manager_finetune = ParamsManager(param_file, key_read="FineTune")

    def add_model(self, model):
        self.model = model

    def KNearestNeighbors(self):
        """
        """
        self.KNN = KNeighborsClassifier()
        return self.KNN

    def NaiveBayes(self):
        """
        Naive Bayes assumes the data to be normally distributed which can be
        achieved by scaling using the MaxAbsScaler.
        """
        self.NB = GaussianNB()
        return self.NB

    def RandomForestClassifier(self):
        """
        n_jobs: Parelizacion en la computacion.
        oob_score: True, muestreo aleatorio.
        n_estimadores = numero de arboles en el bosque
        max_features = numero maximo de caracteristicas consideradas para dividir un nodo
        max_depth = numero maximo de niveles en cada arbol de decision
        min_samples_split = numero minimo de puntos de datos colocados en un nodo antes de que el nodo se divida
        min_samples_leaf = numero minimo de puntos de datos permitidos en un nodo hoja
        bootstrap = metodo para muestrear puntos de datos (con o sin reemplazo)

        """
        self.modelRF = RandomForestClassifier(
            n_estimators=self.manager_models.get_params()["RandomForestClassifier"]["n_estimators"],
            criterion=self.manager_models.get_params()["RandomForestClassifier"]["criterion"],
            max_depth=self.manager_models.get_params()["RandomForestClassifier"]["max_depth"],
            min_samples_split=self.manager_models.get_params()["RandomForestClassifier"]["min_samples_split"],
            min_samples_leaf=self.manager_models.get_params()["RandomForestClassifier"]["min_samples_leaf"],
            min_weight_fraction_leaf=self.manager_models.get_params()["RandomForestClassifier"]["min_weight_fraction_leaf"],
            max_features=self.manager_models.get_params()["RandomForestClassifier"]["max_features"],
            min_impurity_decrease=self.manager_models.get_params()["RandomForestClassifier"]["min_impurity_decrease"],
            bootstrap=self.manager_models.get_params()["RandomForestClassifier"]["bootstrap"],
            oob_score=self.manager_models.get_params()["RandomForestClassifier"]["oob_score"],
            n_jobs=self.manager_models.get_params()["RandomForestClassifier"]["n_jobs"],
            random_state=self.random_state,
            verbose=self.manager_models.get_params()["RandomForestClassifier"]["verbose"],
            warm_start=self.manager_models.get_params()["RandomForestClassifier"]["warm_start"],
            # ccp_alpha=self.manager_models.get_params()["RandomForestClassifier"]["ccp_alpha"]
        )
        return self.modelRF

    def AdaBoostClassifier(self, **params):
        from sklearn.ensemble import AdaBoostClassifier
        return AdaBoostClassifier(random_state=self.random_state, **params)

    def GradientBoostingClassifier(self, **params):
        from sklearn.ensemble import GradientBoostingClassifier
        return GradientBoostingClassifier(random_state=self.random_state, **params)

    def ExtraTreesClassifier(self, **params):
        from sklearn.ensemble import ExtraTreesClassifier
        return ExtraTreesClassifier(random_state=self.random_state, **params)

    def SupportVectorMachine(self, **params):
        """
        """
        self.SVM = SVC(**params)
        return self.SVM

    def XGBoost(self, name="CBT"):
        """
        "min_child_weight": [ Minimum sum of instance weight (hessian) needed in a child.
        "objective": learning task.
        "eval_metric": Evaluation metrics for validation dat.
        "max_depth": Maximum depth of a tree. Increasing this value will make the model more complex and more likely to overfit
        "max_delta_step": /Maximum delta step we allow each leaf output to be. If the value is set to 0, it means there is no constraint.
        "sampling_method": The method to use to sample the training instances.
        "subsample": Subsample ratio of the training instances. Setting it to 0.5 means that XGBoost would randomly sample half of the training data prior to growing trees. and this will prevent overfitting.
        "eta": tep size shrinkage used in update to prevents overfitting.
        "gamma": Minimum loss reduction required to make a further partition on a leaf node of the tree.
        "lambda": L2 regularization term on weights. Increasing this value will make model more conservative.
        "alpha": L1 regularization term on weights. Increasing this value will make model more conservative.
        "tree_method":  he tree construction algorithm used in XGBoost.
        "predictor": The type of predictor algorithm to use.
        "num_parallel_tree": umber of parallel trees constructed during each iteration.

        https://www.analyticsvidhya.com/blog/2016/03/complete-guide-parameter-tuning-xgboost-with-codes-python/
        """
        self.modelXGBoost = modelXGBoost(name=name, random_state=self.random_state)
        return self.modelXGBoost

    def LightBoost(self, name="LBT"):

        self.LBoost = modelLightBoost(name=name, random_state=self.random_state)
        return self.LBoost

    def CatBoost(self, name="CBT"):

        self.CBoost = modelCatBoost(name=name, random_state=self.random_state)
        return self.CBoost

    def append_summary(self, model, X_train, X_test, y_train, y_test, name):
        train_start = time.perf_counter()
        score, _, _ = self.eval_Kfold_CV(model, X_train, X_test, y_train, y_test,
                                            n_splits=self.n_splits, shuffle=True, mute=True)
        train_end = time.perf_counter()
        prediction_start = time.perf_counter()
        _ = model.predict(X_test)
        prediction_end = time.perf_counter()

        self.names.append(name)
        self.utrain.append(train_end-train_start)
        self.utimes.append(prediction_end-prediction_start)

        return score

    def Pipeline_SelectModel(self, X, y, n_splits=5, select="XGBoost"):

        # Lista de modelos a optimizar
        self.scores = []
        self.names = []
        self.utrain = []
        self.utimes = []
        self.n_splits = n_splits
        self.features = X.columns.tolist()

        # Conjunto de entrenamiento y de test
        X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2, random_state=self.random_state)

        # KNearestNeighbors
        if self.manager_models.get_params()["select_models"]["KNN"]:
            _model = self.KNearestNeighbors()
            score = self.append_summary(_model, X_train, X_test, y_train, y_test, "KNearestNeighbors")
            self.scores.append(score)
            if select == "KNearestNeighbors":
                self.add_model(_model)
            print("Modelo: KNearestNeighbors --> Mean Accuracy: {:.3f}%\n".format(score*100))

        # NaiveBayes
        if self.manager_models.get_params()["select_models"]["NaiveBayes"]:
            from sklearn.preprocessing import MaxAbsScaler
            _model = self.NaiveBayes()
            scaler_gnb = MaxAbsScaler()
            sdss_train = scaler_gnb.fit_transform(X_train)
            sdss_test = scaler_gnb.fit_transform(X_test)
            pd_sdss_train = pd.DataFrame(columns=X_train.columns.tolist(), data=sdss_train)
            pd_sdss_test = pd.DataFrame(columns=X_test.columns.tolist(), data=sdss_test)
            score = self.append_summary(_model, pd_sdss_train, pd_sdss_test, y_train, y_test, "NaiveBayes")
            self.scores.append(score)
            if select == "NaiveBayes":
                self.add_model(_model)
            print("Modelo: NaiveBayes --> Mean Accuracy: {:.3f}%\n".format(score*100))

        # SupportVectorMachine
        if self.manager_models.get_params()["select_models"]["SVM"]:
            _model = self.SupportVectorMachine()
            score = self.append_summary(_model, X_train, X_test, y_train, y_test, "SupportVectorMachine")
            self.scores.append(score)
            if select == "SupportVectorMachine":
                self.add_model(_model)

            print("Modelo: SupportVectorMachine --> Mean Accuracy: {:.3f}%\n".format(score*100))

        # RandomForestClassifier
        if self.manager_models.get_params()["select_models"]["RandomForestClassifier"]:
            _model = self.RandomForestClassifier()
            score = self.append_summary(_model, X_train, X_test, y_train, y_test, "RandomForestClassifier")
            self.scores.append(score)
            if select == "RandomForestClassifier":
                self.add_model(_model)
            print("Modelo: RandomForestClassifier --> Mean Accuracy: {:.3f}%\n".format(score*100))

        # XGBoost
        if self.manager_models.get_params()["select_models"]["XGBoost"]:
            _model = self.XGBoost(name="XGBoost")
            _model.fit(X, y, verbose=0, mute=True)
            # score = self.append_summary(_model, X_train, X_test, y_train, y_test, "XGBoost")
            train_start = time.perf_counter()
            score, _ = _model.fit_cv(X, y, nfold=n_splits, verbose=0)
            self.scores.append(score)
            train_end = time.perf_counter()

            prediction_start = time.perf_counter()
            _ = _model.predict(X_test)
            prediction_end = time.perf_counter()

            if select == "XGBoost":
                self.add_model(_model)

            print("Modelo: XGBoost --> Mean Accuracy: {:.3f}%\n".format(score*100))

            self.names.append("XGBoost")
            self.utrain.append(train_end-train_start)
            self.utimes.append(prediction_end-prediction_start)

        # LightGBM
        if self.manager_models.get_params()["select_models"]["LightGBM"]:

            _model = self.LightBoost(name="LBT")
            _model.fit(X, y, verbose=0, mute=True)
            train_start = time.perf_counter()
            score, _ = _model.fit_cv(X, y, nfold=n_splits, verbose=0)
            self.scores.append(score)
            train_end = time.perf_counter()

            prediction_start = time.perf_counter()
            _ = _model.predict(X_test)
            prediction_end = time.perf_counter()

            if select == "LightGBM":
                self.add_model(_model.model)

            print("Modelo: LightGBM --> Mean Accuracy: {:.3f}%\n".format(score*100))

            self.names.append("LightGBM")
            self.utrain.append(train_end-train_start)
            self.utimes.append(prediction_end-prediction_start)

        # CatBoost
        if self.manager_models.get_params()["select_models"]["CatBoost"]:
            _model = self.CatBoost(name="CBT")
            train_start = time.perf_counter()
            score = _model.fit_cv(X, y, fold_count=self.n_splits, shuffle=True, stratified=True,
                                    plot=False, verbose=0)
            self.scores.append(np.mean(score["test-Accuracy-mean"]))
            train_end = time.perf_counter()

            _model.fit(X, y, plot=False, verbose=0)
            prediction_start = time.perf_counter()
            _model.model.predict(_model.eval_data)
            prediction_end = time.perf_counter()

            if select == "CatBoost":
                self.add_model(_model.model)

            print("Modelo: CatBoost --> Mean Accuracy: {:.3f}%\n".format(np.mean(score["test-Accuracy-mean"])*100))

            self.names.append("CatBoost")
            self.utrain.append(train_end-train_start)
            self.utimes.append(prediction_end-prediction_start)

        resultados = pd.DataFrame({"Modelo": self.names, "Mean Accuracy": self.scores,
                                    "Tiempo Entrenamiento": self.utrain, "Tiempo Prediccion": self.utimes})

        return resultados

    def Pipeline_SelectEmsembleModel(self, X, y, n_splits=10, mute=False, scoring="accuracy",
                                        display=True, save_image=False, path="/", AB=True):

        # X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2, random_state=self.random_state)
        X_train, y_train = X, y
        ensembles = []
        if AB:
            ensembles.append(('AB', self.AdaBoostClassifier()))
        ensembles.append(('GBM', self.GradientBoostingClassifier()))
        ensembles.append(('ET', self.ExtraTreesClassifier()))
        ensembles.append(('RF', RandomForestClassifier(random_state=self.random_state)))
        ensembles.append(('XGB', XGBClassifier(random_state=self.random_state)))
        ensembles.append(('LGBM', LGBMClassifier(random_state=self.random_state)))

        results = []
        names = []
        for name, model in ensembles:
            kfold = StratifiedKFold(n_splits=n_splits, random_state=self.random_state)
            cv_results = cross_val_score(model, X_train, y_train, cv=kfold, scoring=scoring)
            results.append(cv_results)
            names.append(name)
            msg = "%s: %f (%f)" % (name, cv_results.mean(), cv_results.std())
            # resultados = zip(name, results)
            if not mute:
                print(msg)

        scores = pd.DataFrame(np.asarray(results).T, columns=names)

        if display:
            figure, axs = plt.subplots(1, 2, figsize=(16, 5))
            ax = axs.flatten()

            # Compare Algorithms
            ax[0].set_title('Ensemble Algorithm Comparison')
            ax[0].boxplot(results)
            ax[0].set_xticklabels(names)
            if AB:
                axis = ["AB", "GBM", "ET", "RF", "XGB", "LGM"]
            else:
                axis = ["GBM", "ET", "RF", "XGB", "LGM"]
            scores_mean = np.mean(scores, axis=0)
            scores_std = np.std(scores, axis=0)
            ax[1].grid()
            ax[1].fill_between(axis, scores_mean - scores_std,
                                scores_mean + scores_std, alpha=0.1,
                                color="r")
            ax[1].plot(axis, scores_mean, 'o-', color="r",
                         label="CV score")
            ax[1].legend(loc="best")
            ax[1].set_title('Cross-validation score')
            figure.tight_layout()
            plt.show()

            if save_image:
                plt.savefig(path)

        return scores

    def Pipeline_FeatureSelect(self, X, y, n_splits=10, mute=False, scoring="accuracy", n_features=20,
                                display=True, save_image=False, path="/"):

        # X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2, random_state=self.random_state)
        X_train, y_train = X, y
        models = []
        models.append(('GBM', self.GradientBoostingClassifier()))
        models.append(('ET', self.ExtraTreesClassifier()))
        models.append(('RF', RandomForestClassifier(random_state=self.random_state)))
        models.append(('XGB', XGBClassifier(random_state=self.random_state)))
        models.append(('LGBM', LGBMClassifier(random_state=self.random_state)))

        results = []
        names = []
        df = pd.DataFrame()

        for name, model in models:
            if not mute:
                print("modelo: {}".format(name))
            if not mute:
                print(".... Fitting")
            model.fit(X_train, y_train)
            if not mute:
                print(".... Permutation importance")
            result = permutation_importance(model, X_train, y_train, n_repeats=10,
                                            random_state=99)
            tree_importance_sorted_idx = np.argsort(model.feature_importances_)
            _ = np.arange(0, len(model.feature_importances_)) + 0.5

            name_features = "features_" + str(name)
            imp_features = "importance" + str(name)
            df[name_features] = X.columns[tree_importance_sorted_idx]
            df[imp_features] = model.feature_importances_[tree_importance_sorted_idx]

            features = df[name_features].values.tolist()[-n_features:]
            _X_train = X_train[features]
            _y_train = y_train
            if not mute:
                print(".... Select Features:")
                print(features)

            if not mute:
                print(".... Cross Validation")
            kfold = StratifiedKFold(n_splits=n_splits, random_state=99)
            cv_results = cross_val_score(model, _X_train, _y_train, cv=kfold, scoring=scoring)
            results.append(cv_results)
            names.append(name)
            msg = "%s: %f (%f)" % (name, cv_results.mean(), cv_results.std())
            # resultados = zip(name, results)
            if not mute:
                print(".... Append Results:")
                print(msg)
                print("\n")

        scores = pd.DataFrame(np.asarray(results).T, columns=names)

        if display:
            figure, axs = plt.subplots(1, 2, figsize=(16, 5))
            ax = axs.flatten()

            # Compare Algorithms
            ax[0].set_title('Algorithm Comparison')
            ax[0].boxplot(results)
            ax[0].set_xticklabels(names)
            axis = ["GBM", "ET", "RF", "XGB", "LGM"]

            scores_mean = np.mean(scores, axis=0)
            scores_std = np.std(scores, axis=0)
            ax[1].grid()
            ax[1].fill_between(axis, scores_mean - scores_std,
                                scores_mean + scores_std, alpha=0.1,
                                color="r")
            ax[1].plot(axis, scores_mean, 'o-', color="r",
                         label="CV score")
            ax[1].legend(loc="best")
            ax[1].set_title('Cross-validation score')
            figure.tight_layout()
            plt.show()

            if save_image:
                plt.savefig(path)

        return scores, df

    def Pipeline_StackingClassifier(self, X, y, n_splits=5):
        # Lista de modelos
        self.models = []

        # KNearestNeighbors
        if self.manager_models.get_params()["stacking_models"]["KNN"]:
            _model = self.KNearestNeighbors()
            self.models.append(("KNearestNeighbors", _model))
        # NaiveBayes
        if self.manager_models.get_params()["stacking_models"]["NaiveBayes"]:
            _model = self.NaiveBayes()
            self.models.append(("NaiveBayes", _model))
        # SupportVectorMachine
        if self.manager_models.get_params()["stacking_models"]["SVM"]:
            _model = self.SupportVectorMachine()
            self.models.append(("SupportVectorMachine", _model))
        # RandomForestClassifier
        if self.manager_models.get_params()["stacking_models"]["RandomForestClassifier"]:
            _model = self.RandomForestClassifier()
            self.models.append(("RandomForestClassifier", _model))
        # XGBoost
        if self.manager_models.get_params()["stacking_models"]["XGBoost"]:
            _model = self.XGBoost(name="XGBoost")
            self.models.append(("XGBoost", _model))
        # LightGBM
        if self.manager_models.get_params()["stacking_models"]["LightGBM"]:
            _model = self.LightBoost(name="LBT")
            self.models.append(("LightGBM", _model))
        # CatBoost
        if self.manager_models.get_params()["stacking_models"]["CatBoost"]:
            _model = self.CatBoost(name="CBT")
            self.models.append(("CatBoost", _model))

    def _cv_results(self, X_train, Y_train, model, kfold, name, verbose=1):
        cv_results = cross_val_score(model, X_train, Y_train, cv=kfold, scoring='accuracy')
        msg = "%s: %f (%f)" % (name, cv_results.mean(), cv_results.std())
        if verbose > 0:
            print(msg)
        return cv_results

    def Ablacion_relativa(self, pipeline, X, y, n_splits=10, mute=False, std=True, scoring="accuracy",
                            display=True, save_image=False, path="/"):
        kfold = StratifiedKFold(n_splits=n_splits, random_state=99)

        models = []
        models.append(('AB', self.AdaBoostClassifier()))
        models.append(('GBM', self.GradientBoostingClassifier()))
        models.append(('RF', RandomForestClassifier(random_state=self.random_state)))
        models.append(('ET', self.ExtraTreesClassifier()))
        models.append(('LGM', LGBMClassifier(random_state=self.random_state)))
        models.append(('XGB', XGBClassifier(random_state=self.random_state)))
        models.append(('SVM', self.SupportVectorMachine()))
        models.append(('KNN', self.KNearestNeighbors()))

        scores_mean = []
        scores_std = []
        names_models = []

        for name_model, model in models:

            names = []
            results = []
            name = "Inicial"

            if not mute:
                print("\n", name_model)

            resu = self._cv_results(X, y, model, kfold, name)
            results.append(resu)
            names.append(name)

            for name, transf in pipeline:
                X_train = transf.fit_transform(X, y)
                Y_train = y
                resu = self._cv_results(X_train, Y_train, model, kfold, name)
                results.append(resu)
                names.append(name)

            scores = pd.DataFrame(np.asarray(results).T, columns=names)
            scores_mean.append(np.mean(scores, axis=0))
            scores_std.append(np.std(scores, axis=0))
            names_models.append(name_model)

        if display:
            fig, ax = plt.subplots(figsize=(14, 6))

            for i in range(len(scores_mean)):
                valor = scores_mean[i]-scores_mean[i].iloc[0]
                if std:
                    ax.fill_between(names, valor - scores_std[i],
                                    valor + scores_std[i], alpha=0.1)
                ax.plot(names, valor, 'o-', label=names_models[i], alpha=0.9)

            ax.plot(names, np.zeros(len(names)), 'ro--', label="cero", alpha=0.9)

            ax.grid()
            ax.legend(loc="best")
            ax.set_title('Mejoras relativas al modelo Inicial')
            fig.tight_layout()
            fig.show()

        if save_image:
            plt.savefig(path)

        return scores_mean, scores_std

    def features_importances(self, clf, X, y, display=True, save_image=False, path="/"):
        import seaborn as sns
        X_train, X_test, y_train, y_test = train_test_split(X, y, random_state=99, stratify=y)

        clf.fit(X_train, y_train)
        print("Accuracy on test data: {:.2f}".format(clf.score(X_test, y_test)))

        result = permutation_importance(clf, X_train, y_train, n_repeats=10,
                                        random_state=99)

        tree_importance_sorted_idx = np.argsort(clf.feature_importances_)
        _ = np.arange(0, len(clf.feature_importances_)) + 0.5

        df = pd.DataFrame()
        df["feature"] = X.columns[tree_importance_sorted_idx]
        df["importance"] = clf.feature_importances_[tree_importance_sorted_idx]

        if display:
            _, _ = plt.subplots(figsize=(10, 30))
            sns.barplot(x="importance", y="feature",
                            data=df.sort_values(by="importance", ascending=False))
            plt.title('Features (avg over folds)')
            plt.show()

        if save_image:
            plt.savefig(path)

        return df

    def eval_Kfold_CV(self, model, X, X_test, y, y_test, n_splits=3, shuffle=True, mute=True):

        _Model = Training.add_model(model)

        resultados, score_general_test = _Model.KFold_CrossValidation(model, X, y, n_splits=n_splits,
                                                                        shuffle=shuffle, mute=mute)
        _predictions = model.predict(X_test)
        score = accuracy_score(y_true=y_test, y_pred=_predictions)

        return score, resultados, score_general_test

    def func_acc(self, prob_pred, y_target):

        _y_pred = np.zeros(len(prob_pred))

        for i in range(0, len(prob_pred)):
            _y_pred[i] = int(np.argmax(prob_pred[i]))

        accuracy = accuracy_score(_y_pred, y_target)

        return accuracy

    def pred_binary(self, prob_pred, y_target, th=0.5):
        return accuracy_score(y_target, np.where(prob_pred > th, 1, 0))

    def replace_multiclass(self, targets):

        _unic = targets.unique().tolist()
        _remp = np.arange(0, len(_unic)).tolist()
        return targets.replace(_unic, _remp), _unic

    def Pipeline_GridSearch(self):
        pass


class wrapper_model(BaseEstimator, TransformerMixin):
    """Wrapper for Estimator."""
    def __init__(self, clf, pipeline_preprocess, random_state=99, name="model", select=False):
        self.clf = clf
        self.random_state = random_state
        self.name = name
        self.buid_Pipeline(pipeline_preprocess, select=select)

    def buid_Pipeline(self, pipeline_preprocess, threshold="median", select=True):
        ## Pipeline
        if select:
            self.pipe = Pipeline([
                ("pre-select", Pipeline([
                    ("preprocess", pipeline_preprocess),
                    ("Select features", SelectFromModel(self.clf, threshold=threshold)),
                ])),
                ("model", self.clf)
                ])
        else:
            self.pipe = Pipeline([
                ("pre-select", Pipeline([
                    ("preprocess", pipeline_preprocess),
                ])),
                ("model", self.clf)
                ])

    def build_param_grid(self, param_grid):
        if isinstance(param_grid, dict):
            self.param_grid = param_grid
        else:
            raise TypeError("Invalid type {}".format(type(param_grid)))

    def fit(self, X, y):
        self.pipe = self.pipe.fit(X, y)

    def fit_cv(self, X, y, n_splits=10, scoring="accuracy", shuffle=False):
        kfold = StratifiedKFold(n_splits=n_splits, shuffle=shuffle, random_state=self.random_state) 
        print("Result CV sin optimizar: {:.3f}%  --  Metric: {}".format(
                cross_val_score(self.pipe, X, y, cv=kfold, scoring=scoring).mean()*100, scoring))

    def Grid_model(self, X, y, n_splits=10, scoring="accuracy", Randomized=False, n_iter=20):

        if not hasattr(self, "param_grid"):
            raise NameError("Not buid Param")

        kfold = StratifiedKFold(n_splits=n_splits, random_state=self.random_state)
        if Randomized:
            search = RandomizedSearchCV(self.pipe, self.param_grid, n_jobs=-1, cv=kfold, n_iter=n_iter)
        else:
            search = GridSearchCV(self.pipe, self.param_grid, n_jobs=-1, cv=kfold)
        search.fit(X, y)
        print("Best parameter (CV score=%0.3f):" % search.best_score_)
        print("  ", search.best_params_)
        self.best_estimador = search.best_estimator_
        self.pipe = search.best_estimator_
        return search.best_params_, search.best_estimator_

    def predict(self, X):
        return self.pipe.predict(X)

    def predict_proba(self, X):
        return self.pipe.predict_proba(X)

    def score(self, X, y):
        return self.pipe.score(X, y)

    def line(self):
        print("="*60)
        print("\n")

    def Restore_model(self, filename="checkpoints/model.pkl"):
        self.clf = pickle.load(open(filename, 'rb'))
        return self.clf

    def Restore_Pipeline(self, filename="checkpoints/Pipeline_model.pkl"):
        self.pipe = pickle.load(open(filename, 'rb'))
        return self.pipe

    @classmethod
    def restore_pipeline_v1(cls, filename, random_state=99, name="Pipeline_model"):
        cls.pipe = pickle.load(open(filename, 'rb'))
        return cls(random_state, name)

    def train_test(self, X, y, test_size=0.1):
        return train_test_split(X, y, test_size=test_size, random_state=self.random_state, stratify=y)

    def cuarentena(self, X, y):
        _preds = self.pipe.predict(X)
        df = X["Id_cat"]
        df["Result"] = _preds != y
        del _preds
        return df

    def Evaluation_model(self, X, y, clases=[0, 1], save=True, ROC=True, n_splits=10,
                            path="checkpoints/"):

        X_train, X_test, y_train, y_test = self.train_test(X, y)

        if not hasattr(self, "best_estimador"):
            clf = self.pipe.named_steps['model']
            transf = self.pipe.named_steps['pre-select']
            self.pipe.fit(X_train, y_train)
            y_pred = self.pipe.predict(X_test)
        else:
            clf = self.best_estimador.named_steps['model']
            transf = self.best_estimador.named_steps['pre-select']
            self.best_estimador.fit(X_train, y_train)
            y_pred = self.best_estimador.predict(X_test)

        eva = Training(clf, random_state=self.random_state)
        _X = pd.DataFrame(transf.fit_transform(X, y))

        self.line()
        print("Confusion Matrix:")
        print(eva.confusion_matrix(y_test, y_pred))
        _ = eva.class_report(y_test, y_pred, clases=clases)

        self.line()
        print("ROC: Cross Validation:")
        resultados, score_general_test = eva.KFold_CrossValidation(clf, _X, y, shuffle=True,
                                                                    n_splits=n_splits, ROC=ROC)
        self.line()
        if not hasattr(self, "best_estimador"):
            clf = self.pipe.named_steps['model']
        else:
            clf = self.best_estimador.named_steps['model']
        # th, res, df = eva.evaluacion_rf_2features(clf, _X, y)
        # print("----> Thresholder óptimo: {:.3f}, result: {:.3f}%".format(th, res*100))

        self.save_general(path, X_train, y_train)

    def save_general(self, path, X_train, y_train):
        ## Save Model
        self.line()
        print("Save Model")
        if not os.path.exists(path):
            os.makedirs(path)
            print("** Path creado: ", path)

        filename = "Pipeline_" + self.name + ".pkl"
        filename = os.path.join(path, filename)
        if not hasattr(self, "best_estimador"):
            self.pipe.fit(X_train, y_train)
            pickle.dump(self.pipe, open(filename, 'wb'))
            print("----> Pipeline guardado en ", filename)
        else:
            self.best_estimador.fit(X_train, y_train)
            pickle.dump(self.best_estimador, open(filename, 'wb'))
            print("----> Pipeline guardado en ", filename)
        self.line()

    def Pipeline_train(self, X, y, n_splits=10, Randomized=False, n_iter=20, threshold='median',
                        clases=[0, 1], ROC=True, path="checkpoints/", eval=True, report=False):
        print("="*60)
        print("  Pipeline: ", self.name)
        self.line()

        # Validacion cruzada sin optimizar
        self.fit_cv(X, y, n_splits=n_splits)

        # Optimización
        if hasattr(self, "param_grid"):
            best_params_, best_estimator_ = self.Grid_model(X, y, Randomized=Randomized, n_iter=20)
            best_params_txt = os.path.join(path, "best_params.txt")

            with open(best_params_txt, 'a') as file_txt:
                file_txt.write(str(self.name + ":"))
                file_txt.write("\n")
                file_txt.write(str(best_params_))
                file_txt.write("\n"*2)
                file_txt.close()

        # Evaluación de resultados
        if eval:
            self.Evaluation_model(X, y, clases=clases, n_splits=n_splits, ROC=ROC, path=path)


class wrapper_pipeline(BaseEstimator, TransformerMixin):
    """Wrapper for Estimator."""
    def __init__(self, filename, name="model", random_state=99):
        self.random_state = random_state
        self.name = name
        self.pipe = pickle.load(open(filename, 'rb'))

    def build_param_grid(self, param_grid):
        if isinstance(param_grid, dict):
            self.param_grid = param_grid
        else:
            raise TypeError("Invalid type {}".format(type(param_grid)))

    def fit(self, X, y):
        self.pipe = self.pipe.fit(X, y)

    def fit_cv(self, X, y, n_splits=10, scoring="accuracy", shuffle=False):
        kfold = StratifiedKFold(n_splits=n_splits, shuffle=shuffle, random_state=self.random_state) 
        print("Result CV sin optimizar: {:.3f}%  --  Metric: {}".format(
                cross_val_score(self.pipe, X, y, cv=kfold, scoring=scoring).mean()*100, scoring))

    def Grid_model(self, X, y, n_splits=10, scoring="accuracy", Randomized=False, n_iter=20):

        if not hasattr(self, "param_grid"):
            raise NameError("Not buid Param")

        kfold = StratifiedKFold(n_splits=n_splits, random_state=self.random_state)
        if Randomized:
            search = RandomizedSearchCV(self.pipe, self.param_grid, n_jobs=-1, cv=kfold, n_iter=n_iter)
        else:
            search = GridSearchCV(self.pipe, self.param_grid, n_jobs=-1, cv=kfold)
        search.fit(X, y)
        print("Best parameter (CV score=%0.3f):" % search.best_score_)
        print("  ", search.best_params_)
        self.best_estimador = search.best_estimator_
        self.pipe = search.best_estimator_
        return search.best_params_, search.best_estimator_

    def predict(self, X):
        return self.pipe.predict(X)

    def predict_proba(self, X):
        return self.pipe.predict_proba(X)

    def score(self, X, y):
        return self.pipe.score(X, y)

    def line(self):
        print("="*60)
        print("\n")

    def Restore_model(self, filename="checkpoints/model.pkl"):
        self.clf = pickle.load(open(filename, 'rb'))
        return self.clf

    def Restore_Pipeline(self, filename="checkpoints/Pipeline_model.pkl"):
        self.pipe = pickle.load(open(filename, 'rb'))
        return self.pipe

    def train_test(self, X, y, test_size=0.1):
        return train_test_split(X, y, test_size=test_size, random_state=self.random_state, stratify=y)

    def cuarentena(self, X, y):
        _preds = self.pipe.predict(X)
        df = X["Id_cat"]
        df["Result"] = _preds != y
        del _preds
        return df

    def Evaluation_model(self, X, y, clases=[0, 1], save=True, ROC=True, n_splits=10,
                            path="checkpoints/"):

        X_train, X_test, y_train, y_test = self.train_test(X, y)

        if not hasattr(self, "best_estimador"):
            clf = self.pipe.named_steps['model']
            transf = self.pipe.named_steps['pre-select']
            self.pipe.fit(X_train, y_train)
            y_pred = self.pipe.predict(X_test)
        else:
            clf = self.best_estimador.named_steps['model']
            transf = self.best_estimador.named_steps['pre-select']
            self.best_estimador.fit(X_train, y_train)
            y_pred = self.best_estimador.predict(X_test)

        eva = Training(clf, random_state=self.random_state)
        _X = pd.DataFrame(transf.fit_transform(X, y))

        self.line()
        print("Confusion Matrix:")
        print(eva.confusion_matrix(y_test, y_pred))
        _ = eva.class_report(y_test, y_pred, clases=clases)

        self.line()
        print("ROC: Cross Validation:")
        resultados, score_general_test = eva.KFold_CrossValidation(clf, _X, y, shuffle=True,
                                                                    n_splits=n_splits, ROC=ROC)
        self.line()
        if not hasattr(self, "best_estimador"):
            clf = self.pipe.named_steps['model']
        else:
            clf = self.best_estimador.named_steps['model']
        # th, res, df = eva.evaluacion_rf_2features(clf, _X, y)
        # print("----> Thresholder óptimo: {:.3f}, result: {:.3f}%".format(th, res*100))

        ## Save Model
        self.save_general(path, X_train, y_train)

    def save_general(self, path, X_train, y_train):
        ## Save Model
        self.line()
        print("Save Model")
        if not os.path.exists(path):
            os.makedirs(path)
            print("** Path creado: ", path)

        filename = "Pipeline_" + self.name + ".pkl"
        filename = os.path.join(path, filename)
        if not hasattr(self, "best_estimador"):
            self.pipe.fit(X_train, y_train)
            pickle.dump(self.pipe, open(filename, 'wb'))
            print("----> Pipeline guardado en ", filename)
        else:
            self.best_estimador.fit(X_train, y_train)
            pickle.dump(self.best_estimador, open(filename, 'wb'))
            print("----> Pipeline guardado en ", filename)
        self.line()

    def Pipeline_train(self, X, y, n_splits=10, Randomized=False, n_iter=20, threshold='median',
                        clases=[0, 1], ROC=True, path="checkpoints/", eval=True):
        self.line()
        print("="*60)
        print("  Pipeline: ", self.name)
        self.line()

        # Validacion cruzada sin optimizar
        self.fit_cv(X, y, n_splits=n_splits)

        # Optimización
        if hasattr(self, "param_grid"):
            _, _ = self.Grid_model(X, y, Randomized=Randomized, n_iter=20)

        # Evaluación de resultados
        if eval:
            self.Evaluation_model(X, y, clases=clases, n_splits=n_splits, ROC=ROC, path=path)





"""Jaime Sendra Berenguer-2020.
MLearner Machine Learning Library Extensions
Author:Jaime Sendra Berenguer<www.linkedin.com/in/jaisenbe>
License: MIT
"""

import json


class ParamsManager(object):
    """
    Objeto utilizado para leer los datos de un fichero json de configuracion.

    Parameters
    ----------

        - params_file: fichero .json desde el directorio de trabajo.
        - key_read: argumento "key" del fichero .json sobre el que queremos actuar.

    Este objeto nos proporciona los siguientes metodos:

        - get_params(): obtener parametros a partir del "key" especificado
        - update_params(): actualizacion de los valores de un "key" especificado.
        - export_params(): Escribir parametros en un archivo
    """

    def __init__(self, params_file, key_read="info_json"):
        self.params = json.load(open(params_file, "r"))
        self.key_read = key_read

    def get_params(self):
        return self.params[self.key_read]

    def update_params(self, **kwargs):
        for key, value in kwargs.items():
            if key in self.get_params().keys():
                self.params[self.key_read][key] = value

    def export_params(self, filename):
        with open(filename, "w") as f:
            json.dump(
                self.params[self.key_read],
                f,
                ident=4,
                separators=(",", ":"),
                sort_keys=True,
            )
            f.write("\n")




"""Jaime Sendra Berenguer-2020.
MLearner Machine Learning Library Extensions
Author:Jaime Sendra Berenguer<www.linkedin.com/in/jaisenbe>
License: MIT
"""

import chart_studio.plotly as py
import plotly.graph_objects as go
import chart_studio
import matplotlib.pyplot as plt

from mlearner import preprocessing

from sklearn.preprocessing import StandardScaler
from sklearn.decomposition import PCA as sk_pca
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis

chart_studio.tools.set_credentials_file(username='jaisenbe58r', api_key='VKTDdiRq6IFfTHG0dYoZ')


def plotly_histogram2(X, columns, target):
    colors = {2: 'rgb(255,127,20)',
              3: 'rgb(31, 220, 120)',
              13: 'rgb(44, 50, 180)'}
    traces = []
    _targets = sorted(X[target].unique().tolist())

    legend = {2: True, 3: True, 13: True}

    for col in range(2):
        for key in range(len(_targets)):
            traces.append(go.Histogram(x=X[X[target] == _targets[key]][columns[col]], opacity=0.7, 
                                        xaxis="x%s" % (col+1), marker=go.Marker(color=colors[_targets[key]]),
                                        name=_targets[key], showlegend=legend[_targets[key]]))
        legend = {2: False, 3: False, 13: False}

    data = go.Data(traces)
    layout = go.Layout(barmode="overlay",
                        xaxis=go.XAxis(domain=[0, 0.48], title=columns[0]),
                        xaxis2=go.XAxis(domain=[0.52, 1], title=columns[1]),
                        yaxis=go.YAxis(title="Numero de Defectos"),
                        title="Histograma caracteristicas")

    fig = go.Figure(data=data, layout=layout)
    py.iplot(fig)

    return fig


def plot_PCA(data, features):
    X = data[features]
    y = data["categoria"]

    X_std = StandardScaler().fit_transform(X)
    acp = sk_pca(n_components=2)
    Y = acp.fit_transform(X_std)

    results = []

    for name in (2, 3, 13):
        result = go.Scatter(x=Y[y == name, 0], y=Y[y == name, 1],
                            mode="markers", name=name,
                            marker=go.Marker(size=8, line=go.Line(color="rgba(225,225,225,0.2)", width=0.5),
                                                opacity=0.75))
        results.append(result)

    data = go.Data(results)
    layout = go.Layout(xaxis=go.XAxis(title="CP1", showline=False), 
                        yaxis=go.YAxis(title="CP2", showline=False))

    fig = go.Figure(data=data, layout=layout)
    py.iplot(fig)

    return fig


def plot_LDA(data, features):
    X = data[features]
    y = data["categoria"]

    X_std = StandardScaler().fit_transform(X)
    LDA = LinearDiscriminantAnalysis()
    Y = LDA.fit_transform(X_std, y)

    results = []

    for name in (2, 3, 13):
        result = go.Scatter(x=Y[y == name, 0], y=Y[y == name, 1],
                            mode="markers", name=name,
                            marker=go.Marker(size=8, line=go.Line(color="rgba(225,225,225,0.2)", width=0.5),
                                                opacity=0.75))
        results.append(result)

    data = go.Data(results)
    layout = go.Layout(xaxis=go.XAxis(title="CP1", showline=False), 
                        yaxis=go.YAxis(title="CP2", showline=False))

    fig = go.Figure(data=data, layout=layout)
    py.iplot(fig)

    return fig


def FeatureAnalyst(X, feature, target, targets=[2, 3, 13]):
    """Analisis de la caracteristica respecto a las categorias."""
    _targets = sorted(X[target].unique())
    data = preprocessing.ExtractCategories(targets, target=["categoria"]).fit_transform(X)

    figure, axs = plt.subplots(1, 2, figsize=(16, 5))
    ax = axs.flatten()

    # Histograma - Datos brutos
    _data = data
    for i in _targets:
        ax[0].hist(_data[_data[target] == i][feature], alpha=0.7, density=True)
    ax[0].set_title("Histograma - Datos brutos")
    ax[0].legend(_targets)

    # Histograma - Escala Logaritmica
    _data = preprocessing.FixSkewness(columns=[feature], drop=False).fit_transform(data)
    for i in _targets:
        ax[1].hist(_data[_data[target] == i][feature], alpha=0.7, density=True)
    ax[1].set_title("Histograma - Escala Logaritmica")
    ax[1].legend(_targets)



"""
Jaime Sendra Berenguer-2020.
MLearner Machine Learning Library Extensions
Author:Jaime Sendra Berenguer<www.linkedin.com/in/jaisenbe>
License: MIT
"""

import chart_studio
from .ml_plotly import plotly_histogram2, plot_PCA, plot_LDA, FeatureAnalyst

chart_studio.tools.set_credentials_file(username='jaisenbe58r', api_key='VKTDdiRq6IFfTHG0dYoZ')


__all__ = ["plotly_histogram2", "plot_PCA", "plot_LDA", "FeatureAnalyst"]



"""
Jaime Sendra Berenguer-2020.
MLearner Machine Learning Library Extensions
Author:Jaime Sendra Berenguer<www.linkedin.com/in/jaisenbe>
License: MIT
"""

import matplotlib.pyplot as plt
import time
import numpy as np
import os
import pandas as pd
import pickle

from sklearn.model_selection import StratifiedKFold, train_test_split, cross_val_score
from sklearn.metrics import accuracy_score
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.metrics import confusion_matrix, classification_report, roc_curve
from sklearn.model_selection import GridSearchCV, RandomizedSearchCV
from sklearn.pipeline import Pipeline

import tensorflow as tf
from tensorflow.keras.models import load_model


class Neural(BaseEstimator, TransformerMixin):
    def __init__(self, clf, random_state=99, name="Neural_model"):
        self.clf = clf
        self.random_state = random_state
        self.name = name

    @classmethod
    def restore(cls, filename, random_state=99, name="Neural_model"):
        clf = load_model(filename)
        return cls(clf, random_state, name)

    def save_model(self, path, name=None):
        if name is None:
            name_save = self.name + ".h5"
        else:
            name_save = name + ".h5"
        filename = os.path.join(path, name_save)
        self.clf.save(filename)

        return filename

    def summary(self):
        self.clf.summary()

    def fit(self, X, y, **params):
        _history = self.clf.fit(X, y, **params)
        return _history

    def fit_cv(self, X, y, n_splits=10, shuffle=True, random_state=99, mute=False, 
                display=True, **params):

        kfold = StratifiedKFold(n_splits=n_splits, shuffle=True, random_state=random_state)
        cvscores = []
        n_history = []
        models = []
        tempdir = "checkpoints/temp/"
        filename = self.save_model(tempdir)
        i = 0
        for train, test in kfold.split(X, y):
            # clone model
            _model = load_model(filename)
            # fit
            start = time.time()
            history = _model.fit(X[train], y[train], verbose=0, validation_data=(X[test], y[test]), **params)
            n_history.append(history)
            end = time.time() - start
            val_acc = np.mean(history.history["val_accuracy"])*100
            print("Fold %s: %.3f%%,  metric: %s, time: %.3f sec." % (i+1, val_acc,
                                                                        _model.metrics_names[1],
                                                                        end))
            cvscores.append(val_acc)
            models.append(_model)
            i = i+1
        if not mute:
            print("%.2f%% (+/- %.2f%%)" % (np.mean(cvscores), np.std(cvscores)))
        if display:
            self.plot_history(n_history)
        os.remove(filename)

        return models

    def plot_history(self, n_history):
        """Grafica Resultados."""
        fig, axs = plt.subplots(2, 2, figsize=(20, 20))
        ax = axs.flatten()
        fig.suptitle("Train-Validation", fontsize=30)
        i = 0
        for history in n_history:

            loss = history.history["loss"]
            accuracy = history.history["accuracy"]
            val_loss = history.history["val_loss"]
            val_accuracy = history.history["val_accuracy"]

            name = "Fold " + str(i+1)

            ax[0].plot(loss, label=name)
            ax[0].set_title("Train-loss")
            ax[0].legend(loc="upper right")

            ax[1].plot(accuracy, label=name)
            ax[1].set_ylim(0.6, 1.1)
            ax[1].set_title("Train-accuracy")
            ax[1].legend(loc="lower right")

            ax[2].plot(val_loss, label=name)
            ax[2].set_title("Validation-loss")
            ax[2].legend(loc="upper right")

            ax[3].plot(val_accuracy, label=name)
            ax[3].set_ylim(0.6, 1.1)
            ax[3].set_title("Validation-accuracy")
            ax[3].legend(loc="lower right")

            i = i+1
        plt.show()

    def predict(self, X, y=None):
        _predictions = tf.argmax(self.clf.predict(X), axis=1, output_type=tf.int32)
        return _predictions.numpy()

    def predict_proba(self, X, y=None):
        return self.clf.predict(X)

    def score(self, X, y):
        _predictions = tf.argmax(self.predict(X), axis=1, output_type=tf.int32)
        _preds = _predictions.numpy()
        print("Accuracy score: %.3f%%" % (accuracy_score(y, _preds)))
        return accuracy_score(y, _preds)

    def line(self):
        print("="*60)
        print("\n")

    def train_test(self, X, y, test_size=0.1):
        return train_test_split(X, y, test_size=test_size, random_state=self.random_state, stratify=y)

    def cuarentena(self, X, y, IDs):
        _preds = self.predict(X)
        df = pd.DataFrame()
        df["Id_cat"] = IDs
        df["Result"] = _preds != y
        del _preds
        return df

    def _evaluacion_rf_2features(self, data_eval, data_eval_target, save=False, logdir_report="", display=True):
        """
        Funcion que nos selecciona el thresholder mas optimo:

        Inputs:

            - data_eval: Conjunto de evaluacion.
            - probs: Matriz de probabilidades que viene dado por el "lm.predict_proba" del podelo de RL.
            - thresholders: Intervalo de corte para determinar la categoria. viene dado por el "metrics.roc_curve" de sckitlearn.
            - display: si queremos mostrar los datos graficados por pantalla.

        Outputs:

            - th: Thresholder seleccionado.
            - res: El accurancy con ese thresholder.
            - df: Dataframe con el resultado del test.
        """
        results = []

        probs = self.predict_proba(data_eval)
        prob = probs[:, 1]
        df = pd.DataFrame(prob)

        _, _, thresholders = roc_curve(data_eval_target, prob)

        for i in range(len(thresholders)):

            df["prediction"] = np.where(df[0] >= thresholders[i], 1, 0)
            df["actual"] = list(data_eval_target)
            df["result"] = np.where(df["prediction"] == df["actual"], 1, 0)

            resultado = df["result"].sum()/df["result"].count()
            results.append(resultado)

        index = results.index(max(results))
        th = thresholders[index]
        res = max(results)

        if display:

            plt.plot(thresholders, results, marker="o", linestyle="--", color="r")
            x = [th, th]
            y = [0, 1]
            plt.plot(x, y)
            plt.xlabel("thresholders")
            plt.ylabel("Result")
            plt.title("Resultado segun thresholders")
            plt.xlim(-0.1, 1.1)
            plt.show()
            if save:
                plt.savefig(logdir_report + "_eval.png")

        df["prediction"] = np.where(df[0] >= thresholders[index], 1, 0)
        df["actual"] = list(data_eval_target)
        df["result"] = np.where(df["prediction"] == df["actual"], 1, 0)

        return th, res, df

    def Evaluation_model(self, X_train, X_test, y_train, y_test, clases=[0, 1], save=True, n_splits=10, ROC=True,
                            path="checkpoints/", **params):
        self.clf.fit(X_train, y_train, verbose=0, **params)

        y_pred = self.clf.predict(X_test)

        self.line()
        print("Confusion Matrix:")
        print(confusion_matrix(y_test, self.predict(X_test)))
        print("\n")
        print("Classification Report:")
        print(classification_report(y_test, self.predict(X_test), target_names=clases))

        self.line()
        th, res, df = self._evaluacion_rf_2features(X_test, y_test)
        print("----> Thresholder optimo: {:.3f}, result: {:.3f}%".format(th, res*100))

        self.save_general(path, X_train, y_train)

    def save_general(self, path, X_train, y_train):
        ## Save Model
        self.line()
        print("Save Model")
        if not os.path.exists(path):
            os.makedirs(path)
            print("** Path creado: ", path)

        _filename = self.save_model(path, self.name)
        print("----> Modelo guardado en ", _filename)

    def Pipeline_train(self, X, y, X_train, X_test, y_train, y_test, n_splits=10, clases=[0, 1], ROC=True, path="checkpoints/", **params):
        print("="*60)
        print("  Pipeline: ", self.name)
        self.line()

        # Validacion cruzada sin optimizar
        self.fit_cv(X, y, n_splits=n_splits, **params)

        # Evaluacion de resultados
        self.Evaluation_model(X_train, X_test, y_train, y_test, clases=clases, n_splits=n_splits, ROC=ROC, path=path, **params)


class Neural_sklearn(BaseEstimator, TransformerMixin):
    def __init__(self, random_state=99, name="Neural_model_sklearn"):
        self.random_state = random_state
        self.name = name

    def buid_Pipeline(self, fn_clf, pipeline_preprocess, **params):
        """
        Example:
        def fn_clf(optimizer=tf.keras.optimizers.Adam(1e-3),
                 kernel_initializer='glorot_uniform',
                 dropout=0.2):
            model = tf.keras.models.Sequential()
            model.add(tf.keras.Input(shape=(8,)))
            model.add(tf.keras.layers.Dense(16, activation="relu",kernel_initializer=kernel_initializer))
            model.add(tf.keras.layers.Dropout(dropout))
            model.add(tf.keras.layers.Dense(1,activation='sigmoid',kernel_initializer=kernel_initializer))

            model.compile(loss='binary_crossentropy',
                        optimizer=optimizer,
                        metrics=['accuracy'])
            return model

        params:
        {
            nb_epoch:100, 
            batch_size:32, 
            verbose:0
            }
        """
        model = tf.keras.wrappers.scikit_learn.KerasClassifier(build_fn=fn_clf, **params)
        ## Pipeline
        self.pipe = Pipeline([
                ("pre-select", Pipeline([
                    ("preprocess", pipeline_preprocess),
                ])),
                ("model", model)
                ])

    @classmethod
    def restore_pipeline(cls, filename, random_state=99, name="Neural_model_sklearn"):
        cls.pipe = pickle.load(open(filename, 'rb'))
        return cls(random_state, name)

    def build_param_grid(self, param_grid):
        if isinstance(param_grid, dict):
            self.param_grid = param_grid
        else:
            raise TypeError("Invalid type {}".format(type(param_grid)))

    def Grid_model(self, X, y, n_splits=10, scoring="accuracy", Randomized=False, n_iter=20):

        if not hasattr(self, "param_grid"):
            raise NameError("Not buid Param")

        kfold = StratifiedKFold(n_splits=n_splits, random_state=self.random_state)
        if Randomized:
            search = RandomizedSearchCV(self.pipe, self.param_grid, n_jobs=-1, cv=kfold, n_iter=n_iter)
        else:
            search = GridSearchCV(self.pipe, self.param_grid, n_jobs=-1, cv=kfold)
        search.fit(X, y)
        print("Best parameter (CV score=%0.3f):" % search.best_score_)
        print("  ", search.best_params_)
        self.best_estimador = search.best_estimator_
        self.pipe = search.best_estimator_
        return search.best_params_, search.best_estimator_

    def save(self, path, name=None):
        if name is None:
            name_save = self.name
        else:
            name_save = name
        filename = "Pipeline_" + name_save + ".pkl"
        filename = os.path.join(path, filename)
        pickle.dump(self.pipe, open(filename, 'wb'))
        print("----> Pipeline guardado en ", filename)
        return filename

    def save_general(self, path, X_train, y_train):
        ## Save Model
        self.line()
        print("Save Model")
        if not os.path.exists(path):
            os.makedirs(path)
            print("** Path creado: ", path)

        filename = "Pipeline_" + self.name + ".pkl"
        filename = os.path.join(path, filename)

        # self.pipe.fit(X_train, y_train)
        pickle.dump(self.pipe, open(filename, 'wb'))
        print("----> Pipeline guardado en ", filename)
        self.line()

    def summary(self):
        clf = self.pipe.named_steps['model']
        clf.summary()

    def fit(self, X, y):
        _history = self.pipe.fit(X, y)
        return _history

    def fit_cv(self, X, y, n_splits=10, shuffle=True, scoring="accuracy",
                mute=False, display=True):

        kfold = StratifiedKFold(n_splits=n_splits, shuffle=shuffle, random_state=self.random_state)
        print("Result CV sin optimizar: {:.3f}%  --  Metric: {}".format(
                cross_val_score(self.pipe, X, y, cv=kfold, scoring=scoring).mean()*100, scoring))

    def predict(self, X, y=None, binary=True):
        if not binary:
            _predictions = tf.argmax(self.pipe.predict(X), axis=1, output_type=tf.int32).numpy()
        else:
            _predictions = self.pipe.predict(X)
        return _predictions

    def predict_proba(self, X, y=None):
        return self.predict(X)

    def score(self, X, y, binary=True):
        _preds = self.predict(X)
        print("Accuracy score: %.3f%%" % (accuracy_score(y, _preds)))
        return accuracy_score(y, _preds)

    def line(self):
        print("="*60)
        print("\n")

    def train_test(self, X, y, test_size=0.1):
        return train_test_split(X, y, test_size=test_size, random_state=self.random_state, stratify=y)

    def cuarentena(self, X, y, IDs):
        _preds = self.predict(X)
        df = pd.DataFrame()
        df["Id_cat"] = IDs
        df["Result"] = _preds != y
        del _preds
        return df

    def _evaluacion_rf_2features(self, data_eval, data_eval_target, save=False, binary=True,
                                    logdir_report="", display=True):
        """
        Funcion que nos selecciona el thresholder mas optimo:

        Inputs:

            - data_eval: Conjunto de evaluacion.
            - probs: Matriz de probabilidades que viene dado por el "lm.predict_proba" del podelo de RL.
            - thresholders: Intervalo de corte para determinar la categoria. viene dado por el "metrics.roc_curve" de sckitlearn.
            - display: si queremos mostrar los datos graficados por pantalla.

        Outputs:

            - th: Thresholder seleccionado.
            - res: El accurancy con ese thresholder.
            - df: Dataframe con el resultado del test.
        """
        results = []

        probs = self.predict_proba(data_eval)
        if not binary:
            prob = probs[:, 1]
        else:
            prob = probs
        df = pd.DataFrame(prob)

        _, _, thresholders = roc_curve(data_eval_target, prob)

        for i in range(len(thresholders)):

            df["prediction"] = np.where(df[0] >= thresholders[i], 1, 0)
            df["actual"] = list(data_eval_target)
            df["result"] = np.where(df["prediction"] == df["actual"], 1, 0)

            resultado = df["result"].sum()/df["result"].count()
            results.append(resultado)

        index = results.index(max(results))
        th = thresholders[index]
        res = max(results)

        if display:

            plt.plot(thresholders, results, marker="o", linestyle="--", color="r")
            x = [th, th]
            y = [0, 1]
            plt.plot(x, y)
            plt.xlabel("thresholders")
            plt.ylabel("Result")
            plt.title("Resultado segun thresholders")
            plt.xlim(-0.1, 1.1)
            plt.show()
            if save:
                plt.savefig(logdir_report + "_eval.png")

        df["prediction"] = np.where(df[0] >= thresholders[index], 1, 0)
        df["actual"] = list(data_eval_target)
        df["result"] = np.where(df["prediction"] == df["actual"], 1, 0)

        return th, res, df

    def Evaluation_model(self, X, y, clases=[0, 1], save=True, n_splits=10, ROC=True, binary=True,
                            path="checkpoints/", **params):
        X_train, X_test, y_train, y_test = self.train_test(X, y)

        self.pipe.fit(X_train, y_train, **params)
        y_pred = self.predict(X_test)

        self.line()
        print("Confusion Matrix:")
        print(confusion_matrix(y_test, self.predict(X_test)))
        print("\n")
        print("Classification Report:")
        print(classification_report(y_test, self.predict(X_test), target_names=clases))

        if not binary:
            self.line()
            th, res, df = self._evaluacion_rf_2features(X, y, binary=binary)
            print("----> Thresholder optimo: {:.3f}".format(th))


        """
        SIN SOLUCIÓN PARA EL GUARDADO
        """
        ## Save Model
        # self.line()
        # print("Save Model")
        # if not os.path.exists(path):
        #     os.makedirs(path)
        #     print("** Path creado: ", path)

        # filename = "Pipeline_" + self.name + ".pkl"
        # filename = os.path.join(path, filename)
        # if not hasattr(self, "best_estimador"):
        #     self.pipe.fit(X_train, y_train)
        #     pickle.dump(self.pipe, open(filename, 'wb'))
        #     print("----> Pipeline guardado en ", filename)

    def Pipeline_train(self, X, y, n_splits=10, Randomized=False, n_iter=20, threshold='median',
                        clases=[0, 1], ROC=True, path="checkpoints/", eval=True):
        print("="*60)
        print("  Pipeline: ", self.name)
        self.line()

        # Optimizacion
        if hasattr(self, "param_grid"):
            _, _ = self.Grid_model(X, y, Randomized=Randomized, n_iter=n_iter)

        # Validacion cruzada sin optimizar
        self.fit_cv(X, y, n_splits=n_splits)

        # Evaluacion de resultados
        if eval:
            self.Evaluation_model(X, y, clases=clases, n_splits=n_splits, ROC=ROC,
                                    path="checkpoints/")






"""Jaime Sendra Berenguer-2020.
MLearner Machine Learning Library Extensions
Author:Jaime Sendra Berenguer<www.linkedin.com/in/jaisenbe>
License: MIT
"""

import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin
from ..externals.estimator_checks import check_is_fitted


class ReplaceTransformer(BaseEstimator, TransformerMixin):

    """This transformer replace some values with others.

    Attributes
    ----------
    columns: `list` of columns to transformer [n_columns]

    mapping: dict`, for example:
        ```
        mapping = {"yes": 1, "no": 0}
        ```

    Examples
    --------
    For usage examples, please see
    https://jaisenbe58r.github.io/MLearner/user_guide/preprocessing/ReplaceTransformer/

    """
    def __init__(self, columns=None, mapping=None):
        """Init replace missing values."""
        if columns is not None:
            if isinstance(columns, list) or isinstance(columns, tuple):
                self.columns = columns
            else:
                raise TypeError("Invalid type {}".format(type(columns)))
        else:
            self.columns = columns

        if mapping is not None:
            if isinstance(mapping, dict):
                self.mapping = mapping
            else:
                raise TypeError("Invalid type {}".format(type(mapping)))
        else:
            raise NameError("Unspecified mapping")

    def fit(self, X, y=None, **fit_params):
        """Gets the columns to make a replace values.

        Parameters
        ----------
        X : {Dataframe}, shape = [n_samples, n_features]
            Dataframe, where n_samples is the number of samples and
            n_features is the number of features.

        Returns
        --------
        self

        """
        if not isinstance(X, pd.core.frame.DataFrame) and not isinstance(X, pd.core.series.Series):
            raise TypeError("Invalid type {}".format(type(X)))

        if not isinstance(X, pd.core.series.Series):
            _lista = [i for i in self.columns if i not in X.columns.tolist()]
            if len(_lista) > 0:
                raise NameError("The columns {} no exist in Dataframe".format(_lista))

        if self.mapping is not None:
            _keys = list(self.mapping)
            if not isinstance(X, pd.core.series.Series):
                for col in self.columns:
                    _lista = [i for i in _keys if i not in X[col].unique().tolist()]
                    if len(_lista) > 0:
                        raise NameError("The Keys {} no exist on column {}".format(_lista, col))
            else:
                _lista = [i for i in _keys if i not in X.unique().tolist()]
                if len(_lista) > 0:
                    raise NameError("The Keys {} no exist".format(_lista))

        self._fitted = True
        return self

    def transform(self, X):
        """Gets the columns to make a replace values.

        Parameters
        ----------
        X : {Dataframe}, shape = [n_samples, n_features]
            Dataframe of samples, where n_samples is the number of samples and
            n_features is the number of features.

        Returns
        -------
        X_transform : {Dataframe}, shape = [n_samples, n_features]
            A copy of the input Dataframe with the columns replaced.

        """
        check_is_fitted(self, '_fitted')

        if not isinstance(X, pd.core.frame.DataFrame) and not isinstance(X, pd.core.series.Series):
            raise TypeError("Invalid type {}".format(type(X)))

        X_transform = X.copy()
        if not isinstance(X, pd.core.series.Series):
            for f in self.columns:
                X_transform[f] = X_transform[f].replace(self.mapping)
        else:
            X_transform = X_transform.replace(self.mapping)
        return X_transform


class ReplaceMulticlass(BaseEstimator, TransformerMixin):

    """This transformer replace some categorical values with others.

    Attributes
    ----------
    columns: `list` of columns to transformer [n_columns]

    Examples
    --------
    For usage examples, please see
    https://jaisenbe58r.github.io/MLearner/user_guide/preprocessing/ReplaceMulticlass/

    """
    def __init__(self, columns=None):
        """Init replace categorical values."""
        if columns is not None:
            if isinstance(columns, list) or isinstance(columns, tuple):
                self.columns = columns
            else:
                raise TypeError("Invalid type {}".format(type(columns)))
        else:
            self.columns = columns

    def fit(self, X, y=None, **fit_params):
        """Gets the columns to make a replace values.

        Parameters
        ----------
        X : {Dataframe}, shape = [n_samples, n_features]
            Dataframe, where n_samples is the number of samples and
            n_features is the number of features.

        Returns
        --------
        self

        """
        if not isinstance(X, pd.core.frame.DataFrame):
            raise TypeError("Invalid type {}".format(type(X)))

        if self.columns is None:
            self.columns = X.columns

        _lista = [i for i in self.columns if i not in X.columns.tolist()]
        if len(_lista) > 0:
            raise NameError("The columns {} no exist in Dataframe".format(_lista))

        self._fitted = True
        return self

    def transform(self, X):
        """Gets the columns to make a replace to categorical values.

        Parameters
        ----------
        X : {Dataframe}, shape = [n_samples, n_features]
            Dataframe of samples, where n_samples is the number of samples and
            n_features is the number of features.

        Returns
        -------
        X_transform : {Dataframe}, shape = [n_samples, n_features]
            A copy of the input Dataframe with the columns replaced.

        """
        check_is_fitted(self, '_fitted')

        if not isinstance(X, pd.core.frame.DataFrame):
            raise TypeError("Invalid type {}".format(type(X)))

        X_transform = X.copy()
        for f in self.columns:
            _unic = X[f].unique().tolist()
            _remp = np.arange(0, len(_unic)).tolist()
            X_transform[f] = X_transform[f].replace(_unic, _remp)

        return X_transform




"""Jaime Sendra Berenguer-2020.
MLearner Machine Learning Library Extensions
Author:Jaime Sendra Berenguer<www.linkedin.com/in/jaisenbe>
License: MIT
"""

import pandas as pd
import numpy as np

from sklearn.feature_selection import SelectKBest
from sklearn.feature_selection import chi2
from sklearn.feature_selection import RFE
from sklearn.feature_selection import SelectFromModel

from sklearn.ensemble import RandomForestClassifier
from sklearn.linear_model import LogisticRegression
from sklearn.preprocessing import MinMaxScaler

from mlearner.utils import ParamsManager
# from mlearner.models import modelLightBoost

import warnings
warnings.filterwarnings("ignore")

param_file = "mlearner/classifier/config/models.json"


class FeatureSelection(object):
    def __init__(self, random_state=99):
        """
        Inicializacion e la clase de Seleccion de rasgos
        """
        self.random_state = random_state
        self.manager_models = ParamsManager(param_file, key_read="Models")

    def cor_pearson(self, X, y, k='all'):
        """
        Pearson Correlation
            Normalization: no
            Impute missing values: yes
        """
        if not k == 'all':
            if k > len(X.columns.tolist()):
                raise NameError("Numero de features seleccionas (k) mayor a features totales")

        feature_name = X.columns.tolist()
        cor_list = []
        # calculate the correlation with y for each feature
        for i in X.columns.tolist():
            cor = np.corrcoef(X[i], y)[0, 1]
            cor_list.append(cor)
        # replace NaN with 0
        cor_list = [0 if np.isnan(i) else i for i in cor_list]
        # feature name
        if k == 'all':
            cor_feature = X.iloc[:, np.argsort(np.abs(cor_list))[:]].columns.tolist()

        else:
            it = k*-1
            cor_feature = X.iloc[:, np.argsort(np.abs(cor_list))[it:]].columns.tolist()

        # feature selection? 0 for not select, 1 for select
        cor_support = [True if i in cor_feature else False for i in feature_name]

        print(str(len(cor_feature)), 'selected features')

        return cor_support, cor_feature

    def chi2(self, X, y, k='all'):
        """
        Chi-2
            Normalization: MinMaxScaler (values should be bigger than 0)
            Impute missing values: yes
        """
        if not k == 'all':
            if k > len(X.columns.tolist()):
                raise NameError("Numero de features seleccionas (k) mayor a features totales")

        feature_name = X.columns.tolist()
        X_norm = MinMaxScaler().fit_transform(X)
        chi_selector = SelectKBest(chi2, k=k)
        chi_selector.fit(X_norm, y)

        _chi_support = chi_selector.get_support()
        chi_feature = X.loc[:, _chi_support].columns.tolist()
        print(str(len(chi_feature)), 'selected features')

        chi_support = [True if i in chi_feature else False for i in feature_name]

        return chi_support, chi_feature, chi_selector

    def wrapper(self, X, y, k='all'):
        """
        Wrapper
        documentation for RFE: http://scikit-learn.org/stable/modules/generated/sklearn.feature_selection.RFE.html

            Normalization: depend on the used model; yes for LR
            Impute missing values: depend on the used model; yes for LR
        """
        X_norm = MinMaxScaler().fit_transform(X)

        if not k == 'all':
            if k > len(X.columns.tolist()):
                raise NameError("Numero de features seleccionas (k) mayor a features totales")
        else:
            k = len(X.columns.tolist())

        rfe_selector = RFE(estimator=LogisticRegression(), n_features_to_select=k, step=10, verbose=0)
        rfe_selector.fit(X_norm, y)

        rfe_support = rfe_selector.get_support()
        rfe_feature = X.loc[:, rfe_support].columns.tolist()
        print(str(len(rfe_feature)), 'selected features')

        return rfe_support, rfe_feature, rfe_selector

    def embeded(self, X, y):
        """
        Embeded
            documentation for SelectFromModel: http://scikit-learn.org/stable/modules/generated/sklearn.feature_selection.SelectFromModel.html ### 3.1 Logistics Regression L1 Note
            Normalization: Yes
            Impute missing values: Yes
        """
        X_norm = MinMaxScaler().fit_transform(X)

        embeded_lr_selector = SelectFromModel(LogisticRegression(penalty="l2"), '1.25*median')
        embeded_lr_selector.fit(X_norm, y)

        embeded_lr_support = embeded_lr_selector.get_support()
        embeded_lr_feature = X.loc[:, embeded_lr_support].columns.tolist()
        print(str(len(embeded_lr_feature)), 'selected features')

        return embeded_lr_support, embeded_lr_feature, embeded_lr_selector

    def RandomForest(self, X, y, n_estimators=100):
        """
        Random Forest

            Normalization: No
            Impute missing values: Yes
        """
        embeded_rf_selector = SelectFromModel(RandomForestClassifier(n_estimators=n_estimators), threshold='1.25*median')
        embeded_rf_selector.fit(X, y)

        embeded_rf_support = embeded_rf_selector.get_support()
        embeded_rf_feature = X.loc[:, embeded_rf_support].columns.tolist()
        print(str(len(embeded_rf_feature)), 'selected features')

        return embeded_rf_support, embeded_rf_feature, embeded_rf_selector

    def LightGBM(self, X, y, n_estimators=100):
        """
        LightGBM

            Normalization: No
            Impute missing values: No
        """
        from lightgbm import LGBMClassifier
        lgbc = LGBMClassifier(n_estimators=n_estimators)
        # lgbc = modelLightBoost()

        embeded_lgb_selector = SelectFromModel(lgbc, threshold='1.25*median')
        embeded_lgb_selector.fit(X, y)

        embeded_lgb_support = embeded_lgb_selector.get_support()
        embeded_lgb_feature = X.loc[:, embeded_lgb_support].columns.tolist()
        print(str(len(embeded_lgb_feature)), 'selected features')

        return embeded_lgb_support, embeded_lgb_feature, embeded_lgb_selector

    def transform_data(self, X, selector, features):
        """
        Transformar el conjunto de entrenamiento al nuevo esquema proporcionado por el selector
        """
        dff = selector.transform(X)
        df_X = pd.DataFrame(columns=features, data=dff)

        return df_X

    def Summary(self, X, y, k="all", cor_pearson=True, chi2=True, wrapper=True, embeded=True,
                RandomForest=True, LightGBM=True):
        """
        Resumen de la seleccion de caracteristicas.
        """
        feature_name = X.columns.tolist()
        pd.set_option('display.max_rows', None)

        # put all selection together
        feature_selection_df = pd.DataFrame({'Feature': feature_name})

        if cor_pearson:
            cor_support, _ = self.cor_pearson(X, y, k)
            feature_selection_df['Pearson'] = cor_support
        if chi2:
            chi_support, _, _ = self.chi2(X, y, k)
            feature_selection_df['Chi2'] = chi_support
        if wrapper:
            rfe_support, _, _ = self.wrapper(X, y, k)
            feature_selection_df['RFE'] = rfe_support
        if embeded:
            embeded_lr_support, _, _ = self.embeded(X, y)
            feature_selection_df['Logistics'] = embeded_lr_support
        if RandomForest:
            embeded_rf_support, _, _ = self.RandomForest(X, y)
            feature_selection_df['Random Forest'] = embeded_rf_support
        if LightGBM:
            embeded_lgb_support, _, _ = self.LightGBM(X, y)
            feature_selection_df['LightGBM'] = embeded_lgb_support

        # count the selected times for each feature
        feature_selection_df['Total'] = np.sum(feature_selection_df, axis=1)
        # display the top 100
        feature_selection_df = feature_selection_df.sort_values(['Total', 'Feature'], ascending=False)
        feature_selection_df.index = range(1, len(feature_selection_df)+1)
        feature_selection_df.head(100)

        return feature_selection_df




"""Jaime Sendra Berenguer-2020.
MLearner Machine Learning Library Extensions
Author: Jaime Sendra Berenguer<www.linkedin.com/in/jaisenbe>
License: MIT
"""

import pandas as pd


class DataLoad(object):

    """Loading a dataset from a file or dataframe.

    Parameters
    ----------
    data: pandas [n_columns, n_features].
        pandas Dataframe.

    Attributes
    ----------
    n: lenght of dataset.
    start: start iterator.
    end: end iterator.
    num: current iterator.

    Examples
    --------
    For usage examples, please see:
    https://jaisenbe58r.github.io/MLearner/user_guide/load/DataLoad/


    """
    def __init__(self, data):
        self.data = data
        self.n = len(self.data)
        self.start = 0
        self.end = self.n
        self.num = 0

    @classmethod
    def load_data(cls, filename, name="dataset", sep=';', decimal=",", **params):
        """Loading a dataset from a csv file.

        Parameters
        ----------
        filename: `str, path object or file-like object`
            Any valid string path is acceptable. The string could be a URL.
            Valid URL schemes include http, ftp, s3, and file. For file URLs,
            a host is expected. A local file could be:
            `file://localhost/path/to/table.csv`.
            If you want to pass in a path object, pandas accepts any os.PathLike.
            By file-like object, we refer to objects with a read() method,
            such as a file handler (e.g. via builtin open function) or StringIO.

        seps: `str`
            Delimiter to use. If sep is None, the C engine cannot automatically
            detect the separator, but the Python parsing engine can, meaning the
            latter will be used and automatically detect the separator by Python's
            builtin sniffer tool, csv.Sniffer.

        delimiter: `str, default None`
            Alias for sep.

        Attributes
        ----------
        n: lenght of dataset.
        start: start iterator.
        end: end iterator.
        num: current iterator.

        Returns
        -------
        data: Pandas DataFrame, [n_samples, n_classes]
            Dataframe from dataset.

        Examples
        --------
        For usage examples, please see:
        https://jaisenbe58r.github.io/MLearner/user_guide/load/DataLoad/

        """
        data = pd.read_csv(filename, sep=sep, decimal=decimal, **params)
        return cls(data)

    @classmethod
    def load_dataframe(cls, data):
        if not isinstance(data, pd.core.frame.DataFrame):
            raise TypeError("Invalid type {}".format(type(data)))
        return cls(data)

    def reset(self):
        self.num = 0

    def __iter__(self):
        self.num = 0
        return self

    def __next__(self):
        if self.num > self.end:
            raise StopIteration
        else:
            self.num += 1
            return self.data.iloc[self.num - 1].head()

    def __len__(self):
        return self.n



"""
Jaime Sendra Berenguer-2020.
MLearner Machine Learning Library Extensions
Author:Jaime Sendra Berenguer<www.linkedin.com/in/jaisenbe>
License: MIT
"""

from .wine import wine_data
from ._dummy_dataset import data_uniform, data_normal, data_gamma, create_dataset


__all__ = ["wine_data", "data_normal", "data_uniform", "data_gamma", "create_dataset"]



"""
Jaime Sendra Berenguer-2020.
MLearner Machine Learning Library Extensions
Author:Jaime Sendra Berenguer<www.linkedin.com/in/jaisenbe>
License: MIT
"""

from .ml_CatBoost import modelCatBoost
from .ml_LightBoost import modelLightBoost
from .ml_XGBoost import modelXGBoost


__all__ = ["modelCatBoost", "modelLightBoost", "modelXGBoost"]




"""Jaime Sendra Berenguer-2020.
MLearner Machine Learning Library Extensions
Author:Jaime Sendra Berenguer<www.linkedin.com/in/jaisenbe>
License: MIT
"""

import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import os
import datetime
import time
import joblib

from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.model_selection import train_test_split
from sklearn.metrics import accuracy_score

import xgboost as xgb
from xgboost import XGBClassifier

import seaborn as sns

from mlearner.training import Training
from mlearner.utils import ParamsManager

import warnings
warnings.filterwarnings("ignore")

param_file = "mlearner/classifier/config/models.json"


class modelXGBoost(Training, BaseEstimator, ClassifierMixin):
    """
    XGBoost is an optimized distributed gradient boosting library designed to be highly efficient,
    flexible and portable. It implements machine learning algorithms under the Gradient Boosting framework.
    XGBoost provides a parallel tree boosting (also known as GBDT, GBM) that solve many data science
    problems in a fast and accurate way. The same code runs on major distributed environment
    (Hadoop, SGE, MPI) and can solve problems beyond billions of examples.

    Parameters
    ----------
        "min_child_weight": [ Minimum sum of instance weight (hessian) needed in a child.
        "objective": learning task.
        "eval_metric": Evaluation metrics for validation data.
        "max_depth": Maximum depth of a tree. Increasing this value will make the model more complex and more likely to overfit
        "max_delta_step": /Maximum delta step we allow each leaf output to be. If the value is set to 0, it means there is no constraint.
        "sampling_method": The method to use to sample the training instances.
        "subsample": Subsample ratio of the training instances. Setting it to 0.5 means that XGBoost would randomly sample half of the training data prior to growing trees. and this will prevent overfitting.
        "eta": tep size shrinkage used in update to prevents overfitting.
        "gamma": Minimum loss reduction required to make a further partition on a leaf node of the tree.
        "lambda": L2 regularization term on weights. Increasing this value will make model more conservative.
        "alpha": L1 regularization term on weights. Increasing this value will make model more conservative.
        "tree_method":  he tree construction algorithm used in XGBoost.
        "predictor": The type of predictor algorithm to use.
        "num_parallel_tree": umber of parallel trees constructed during each iteration.
        ...

    Documentation
    -------------
        https://xgboost.readthedocs.io/en/latest/
        https://www.analyticsvidhya.com/blog/2016/03/complete-guide-parameter-tuning-xgboost-with-codes-python/
    """
    def __init__(self, name="XGB", random_state=99, train_dir="", params=None, *args, **kwargs):

        self.name = name
        self.train_dir = train_dir + "/" + "model_" + str(self.name) + "/"
        self.random_state = random_state

        if params is None:
            self.get_params_json()
            self.params.update({
                        'model_dir': self.train_dir,
                        "seed": self.random_state})
        else:
            # if isinstance(params)
            self.params = params

        self.model = XGBClassifier(**self.params)
        super().__init__(self.model, random_state=self.random_state)

    def get_params_json(self):
        self.manager_models = ParamsManager(param_file, key_read="Models")
        self.params = self.manager_models.get_params()["XGBoost"]

        self.manager_finetune = ParamsManager(param_file, key_read="FineTune")
        self.params_finetune = self.manager_finetune.get_params()["XGBoost"]

    def dataset(self, X, y, categorical_columns_indices=None, test_size=0.2, *args, **kwarg):

        self.categorical_columns_indices = categorical_columns_indices

        self.X = X
        self.columns = list(X)

        self.y, self.cat_replace = self.replace_multiclass(y)

        self.X_train, self.X_test, self.y_train, self.y_test = train_test_split(self.X, self.y, test_size=test_size, random_state=self.random_state)

        self.dtrain = xgb.DMatrix(self.X_train, label=self.y_train)
        self.dvalid = xgb.DMatrix(self.X_test, label=self.y_test)
        self.all_train_data = xgb.DMatrix(self.X, label=self.y)

    def set_dataset_nosplit(self, X_train, X_test, y_train, y_test, categorical_columns_indices=None, *args, **kwarg):

        self.categorical_columns_indices = categorical_columns_indices

        self.columns = list(X_train)

        _ytrain, _ = self.replace_multiclass(y_train)
        _ytest, _ = self.replace_multiclass(y_test)

        self.X_train = X_train
        self.X_test = X_test
        self.y_train = y_train
        self.y_test = y_test

        self.X = pd.concat([X_train, X_test], axis=0)
        self.y = pd.concat([y_train, y_test], axis=0)

        self.dtrain = xgb.DMatrix(self.X_train, label=self.y_train)
        self.dvalid = xgb.DMatrix(self.X_test, label=self.y_test)
        self.all_train_data = xgb.DMatrix(self.X, label=self.y)

    def replace_multiclass(self, targets):

        _unic = targets.unique().tolist()
        _remp = np.arange(0, len(_unic)).tolist()
        return targets.replace(_unic, _remp), _unic

    def fit(self, X=None, y=None, X_train=None, X_test=None, y_train=None, y_test=None, mute=False,
            use_best_model=True, verbose=0, num_boost_round=100, nosplit=False, **kwargs):

        if not nosplit:
            self.dataset(X, y)
        else:
            self.set_dataset_nosplit(X_train, X_test, y_train, y_test)

        self.params.update({'verbosity': verbose})

        self.model = xgb.train(
                            self.params,
                            self.dtrain,
                            num_boost_round=num_boost_round,
                            # verbosity=verbose,
                            **kwargs)

        _preds = self.model.predict(self.dvalid)
        preds_test = np.where(_preds > 0.5, 1, 0)
        score_test = accuracy_score(self.y_test, preds_test)

        _preds = self.model.predict(self.dtrain)
        preds_train = np.where(_preds > 0.5, 1, 0)
        score_train = accuracy_score(self.y_train, preds_train)

        if not mute:
            print("Accurancy para el conjunto de entrenamiento ---> {:.2f}%".format(score_train*100))
            print("Accurancy para el conjunto de validacion ------> {:.2f}%".format(score_test*100))

    def fit_cv(self, X=None, y=None, X_train=None, X_test=None, y_train=None, y_test=None, num_boost_round=75,
                nfold=5, use_best_model=True, verbose=2, nosplit=False, early_stopping_rounds=75, **kwargs):
        """
        https://xgboost.readthedocs.io/en/latest/parameter.html
        """
        if not nosplit:
            self.dataset(X, y)
        else:
            self.set_dataset_nosplit(X_train, X_test, y_train, y_test)

        self.params.update({'verbose_eval': verbose})
        self.xgb_cv = xgb.cv(
                            self.params,
                            self.all_train_data,
                            num_boost_round,
                            nfold,
                            early_stopping_rounds=early_stopping_rounds,
                            stratified=True,
                            seed=self.random_state)

        loss = "test-" + self.params["metrics"][0]
        optimal_rounds = np.argmin(self.xgb_cv[str(loss) + '-mean'])
        best_cv_score = max(self.xgb_cv[str(loss) + '-mean'])

        if not verbose == 0:
            print("\nOptimal Round: {}\nOptimal Score: {:.3f} + std:{:.3f}".format(
                optimal_rounds, best_cv_score, self.xgb_cv[str(loss) + '-std'][optimal_rounds]))

        results = {"Rounds": optimal_rounds,
                                "Score": best_cv_score,
                                "STDV": self.xgb_cv[str(loss) + '-std'][optimal_rounds],
                                "LB": None,
                                "Parameters": self.params}

        score = self.xgb_cv[str(loss) + '-mean'].mean()
        return score, results

    def func_acc(self, prob_pred, y_target):

        _y_pred = np.zeros(len(prob_pred))

        for i in range(0, len(prob_pred)):
            _y_pred[i] = int(np.argmax(prob_pred[i]))
        accuracy = accuracy_score(_y_pred, y_target)

        return accuracy

    def predict(self, X, *args, **kwargs):
        _X_copy = X.loc[:, self.columns].copy()
        return self.model.predict(xgb.DMatrix(_X_copy), *args, **kwargs)

    def pred_binary(self, X, *args, **kwargs):
        _X_copy = X.loc[:, self.columns].copy()
        preds = self.model.predict(xgb.DMatrix(_X_copy), *args, **kwargs)
        return np.where(preds > 0.5, 1, 0)

    def pred_multiclass(self, X, *args, **kwargs):
        _X_copy = X.loc[:, self.columns].copy()
        return [np.argmax(line) for line in self.model.predict(xgb.DMatrix(_X_copy))]

    def update_model(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self.model, k, v)

    def save_model(self, direct="./checkpoints", name="XGB_model", file_model=".txt"):

        if not os.path.isdir(direct):
            try:
                os.mkdir(direct)
                print("Directorio creado: " + direct)
            except OSError as e:
                raise NameError("Error al crear el directorio")

        current_time = datetime.datetime.now().strftime("%Y%m%d-%H%M%S")

        if file_model == ".txt":
            filename = direct + "/" + name + "_" + current_time + ".txt"
            self.model.save_model(filename)
        elif file_model == ".pkl":
            filename = direct + "/" + name + "_" + current_time + ".pkl"
            joblib.dump(self.model, filename)
        else:
            raise NameError("Type {} not permited".format(file_model))
        print("Modelo guardado en la ruta: " + filename)

    def load_model(self, direct="./checkpoints/XGB_model.txt", file_model=".txt"):

        if not os.path.isdir(direct):
            print("no existe el drectorio especificado")

        if file_model == ".txt":
            self.model = XGBClassifier(model_file=direct)
        elif file_model == ".pkl":
            self.model = joblib.load(direct)
        else:
            raise NameError("Type {} not permited".format(file_model))
        print("Modelo cargado de la ruta: " + direct)

    def index_features(self, features):

        _index = []
        for i in features:
            _index.append(self.X.columns.get_loc(i))

        if _index == []:
            raise NameError("No coincide ninguna de las features introducidas")

        return _index

    def get_important_features(self, display=True, max_num_features=20):

        _model = XGBClassifier()
        _model.fit(self.X, self.y)
        _data = np.array([self.X.columns, _model.feature_importances_])
        _feature_importance_df = pd.DataFrame(_data.T, columns=["Feature Id", "Importances"])
        _feature_importance_df = _feature_importance_df.sort_values(by=['Importances'], ascending=False)

        if display:
            plt.figure(figsize=(12, 6))
            sns.barplot(x="Importances", y="Feature Id", data=_feature_importance_df)
            plt.title('XGBoost features importance:')
        # if display:
        #     xgb.plot_importance(self.model, max_num_features=max_num_features, figsize=(6, 6), title='Feature importance (LightGBM)')
        #     plt.show()

        return _feature_importance_df

    def FineTune_SearchCV(self, X=None, y=None, X_train=None, X_test=None, y_train=None, y_test=None, params=None, params_finetune=None, ROC=False,
                            randomized=True, cv=10, n_iter=10, replace_model=True, verbose=0, nosplit=False, finetune_dir=""):
        self.get_params_json()
        self.finetune_dir = finetune_dir + "/" + "model_finetune_" + str(self.name) + "/"
        self.params.update({
                    'train_dir': self.finetune_dir,
                    "seed": self.random_state})
        if params is not None:
            self.params = params
        if params_finetune is not None:
            self.params_finetune = params_finetune

        if not nosplit:
            self.dataset(X, y)
        else:
            self.set_dataset_nosplit(X_train, X_test, y_train, y_test)

        self.params.update({'verbosity': verbose})
        self.model = XGBClassifier(**self.params)

        self._best_Parameters, self.results_df = self.FineTune(self.model, self.X_train, self.y_train, self.params_finetune,
                                                                randomized=True, cv=cv, n_iter=n_iter, verbose=1)
        self.params.update(**self._best_Parameters)
        self.fit(self.X_train, self.y_train, verbose=1)

        score = accuracy_score(self.y_test, self.pred_multiclass(self.X_test))
        print("Resultado del conjunto de test con los parametros optimos: {:.2f}%".format(score*100))
        print("\n")
        print("Report clasificacion con el conjunto de test: ")
        self.evaluate(self.model, xgb.DMatrix(self.X_test), self.y_test)
        print("\n")
        print("Cross validation con todos los datos del dataset: ")
        print("\n")
        self.KFold_CrossValidation(XGBClassifier(**self._best_Parameters), self.X_test, self.y_test, n_splits=cv, ROC=ROC, shuffle=True, mute=False,
                                    logdir_report="", display=True, save_image=True, verbose=0)

        return self._best_Parameters, self.results_df

    def SeedDiversification_cv(self, X=None, y=None, X_train=None, X_test=None, y_train=None, y_test=None, n_iter=10, n_max=2018-2022, cv=10,
                                nosplit=False, finetuneseed_dir="", display=True, save_image=True, verbose=0):
        allmodelstart = time.time()
        self.get_params_json()

        self.finetune_dir = finetuneseed_dir + "/" + "model_finetune_seed" + str(self.name) + "/"
        self.params.update({'train_dir': self.finetune_dir,
                            'verbosity': verbose})
        if not nosplit:
            self.dataset(X, y)
        else:
            self.set_dataset_nosplit(X_train, X_test, y_train, y_test)

        self.params.update({'verbosity': verbose})
        self.model = XGBClassifier(**self.params)

        _rd = np.random.uniform(0, n_max, n_iter).astype(np.int32).tolist()
        params_finetuneseed = {"seed": _rd}
        del(_rd)

        self._best_Parameters, self.results_df = self.FineTune(self.model, self.X, self.y, params_finetuneseed,
                                                                randomized=False, cv=cv, n_iter=n_iter,
                                                                verbose=1, mute=True)

        print("All Model Runtime: %0.2f Minutes" % ((time.time() - allmodelstart)/60))

        print("Diversificacion de la semilla - mean AUC: {:.2f}% - std AUC: {:.5f}".format(
                    self.results_df['mean_test_AUC'].mean()*100, self.results_df['std_test_AUC'].mean()))

        print("Diversificacion de la semilla - mean Acc: {:.2f}% - std Acc: {:.5f}".format(
                    self.results_df['mean_test_Accuracy'].mean()*100, self.results_df['std_test_Accuracy'].mean()))

        return self._best_Parameters, self.results_df

    def __getattr__(self, attr):
        """
        Pass all other method calls to self.model.
        """
        return getattr(self.model, attr)




"""Jaime Sendra Berenguer-2020.
MLearner Machine Learning Library Extensions
Author:Jaime Sendra Berenguer<www.linkedin.com/in/jaisenbe>
License: MIT
"""

import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import os
import datetime

# from sklearn.base import BaseEstimator, TransformerMixin, ClassifierMixin
from sklearn.model_selection import RandomizedSearchCV, RepeatedStratifiedKFold
from sklearn.model_selection import train_test_split
from sklearn.metrics import make_scorer, accuracy_score, cohen_kappa_score

import catboost
from catboost import CatBoostClassifier
from catboost import MetricVisualizer

import seaborn as sns
import hyperopt
from numpy.random import RandomState
import scipy

from mlearner.utils import ParamsManager

import warnings
warnings.filterwarnings("ignore")

param_file = "mlearner/classifier/config/models.json"


class modelCatBoost(object):

    def __init__(self, name="CBT", random_state=99, *args, **kwargs):

        self.name = name
        self.train_dir = "model_" + str(self.name) + "/"
        self.random_state = random_state

        self.manager_models = ParamsManager(param_file, key_read="Models")
        self.params = self.manager_models.get_params()["CatBoost"]
        self.params.update({
                    'train_dir': self.train_dir,
                    "random_state": self.random_state
                })

        self.model = CatBoostClassifier(**self.params)

    def dataset(self, X, y, categorical_columns_indices=None, test_size=0.2, *args, **kwargs):

        self.categorical_columns_indices = categorical_columns_indices
        self.X = X
        self.columns = list(X)

        self.y, self.cat_replace = self.replace_multiclass(y)
        self.X_train, self.X_test, self.y_train, self.y_test = train_test_split(self.X, self.y, test_size=test_size, random_state=self.random_state)

        self.train_data = catboost.Pool(data=self.X_train.values,
                                        label=self.y_train.values,
                                        cat_features=self.categorical_columns_indices)
        self.eval_data = catboost.Pool(data=self.X_test.values,
                                        label=self.y_test.values,
                                        cat_features=self.categorical_columns_indices)
        self.all_train_data = catboost.Pool(data=self.X.values,
                                            label=self.y.values,
                                            cat_features=self.categorical_columns_indices)

    def replace_multiclass(self, targets):

        _unic = targets.unique().tolist()
        _remp = np.arange(0, len(_unic)).tolist()
        return targets.replace(_unic, _remp), _unic

    def fit(self, X, y, use_best_model=True, plot=True, save_snapshot=False, verbose=0, *args, **kwargs):

        self.dataset(X, y)
        _params = self.model.get_params()

        if verbose:
            _verbose = 0
        else:
            _verbose = _params["verbose"]

        return self.model.fit(self.train_data,
                                verbose=_verbose,
                                eval_set=self.eval_data,
                                use_best_model=use_best_model,
                                plot=plot,
                                save_snapshot=save_snapshot,
                                **kwargs)

        _preds = self.model.predict(self.dvalid)
        preds_test = np.where(_preds > 0.5, 1, 0)
        score_test = accuracy_score(self.y_test, preds_test)

        _preds = self.model.predict(self.dtrain)
        preds_train = np.where(_preds > 0.5, 1, 0)
        score_train = accuracy_score(self.y_train, preds_train)

        if not verbose == 0:
            print("Accurancy para el conjunto de entrenamiento ---> {:.2f}%".format(score_train*100))
            print("Accurancy para el conjunto de validacion ------> {:.2f}%".format(score_test*100))

    def fit_cv(self, X, y, fold_count=4, shuffle=True, stratified=True, plot=True, verbose=100):

        self.dataset(X, y)

        _params = self.model.get_params()
        _params.update({'verbose': verbose})

        _scores = catboost.cv(pool=self.all_train_data,
                                params=_params,
                                fold_count=fold_count,
                                seed=self.random_state,
                                shuffle=shuffle,
                                verbose=verbose,
                                plot=plot)
        if not verbose == 0:
            print('Best validation accuracy score: {:.2f}±{:.2f} on step {}'.format(
        np.max(_scores['test-Accuracy-mean']),
        _scores['test-Accuracy-std'][np.argmax(_scores['test-Accuracy-mean'])],
        np.argmax(_scores['test-Accuracy-mean'])))

        return _scores

    def copy(self, *args, **kwargs):
        returned_classifier = CatBoostClassifier()
        returned_classifier.catboost_classifier = self.model.copy()
        returned_classifier.columns = self.columns
        return returned_classifier

    def update_model(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self.model, k, v)

    def save_model(self, direct="./checkpoints", name="catboost_model"):

        if not os.path.isdir(direct):
            try:
                os.mkdir(direct)
                print("Directorio creado: " + direct)
            except OSError as e:
                raise NameError("Error al crear el directorio")
        current_time = datetime.datetime.now().strftime("%Y%m%d-%H%M%S")
        filename = direct + "/" + name + "_" + current_time + ".dump"
        self.model.save_model(filename)
        print("Modelo guardado en la ruta: " + filename)

    def load_model(self, direct="./checkpoints", name="catboost_model"):

        if not os.path.isdir(direct):
            print("no existe el drectorio especificado")
        filename = direct + "/" + name + ".dump"
        self.model.load_model(filename)
        print("Modelo cargado de la ruta: " + filename)

    def predict(self, X, *args, **kwargs):
        _X_copy = X.loc[:, self.columns].copy()
        return self.model.predict(_X_copy.values, *args, **kwargs)

    def predict_proba(self, X, *args, **kwargs):
        _X_copy = X.loc[:, self.columns].copy()
        return self.model.predict_proba(_X_copy.values, *args, **kwargs)

    def add_cat_features(self, index_features):

        self.categorical_columns_indices = index_features
        print(self.categorical_columns_indices)

        self.train_data = catboost.Pool(data=self.X_train,
                                        label=self.y_train,
                                        cat_features=self.categorical_columns_indices)
        self.eval_data = catboost.Pool(data=self.X_test,
                                        label=self.y_test,
                                        cat_features=self.categorical_columns_indices)
        self.all_train_data = catboost.Pool(data=self.X,
                                            label=self.y,
                                            cat_features=self.categorical_columns_indices)

    def index_features(self, features):

        _index = []
        for i in features:
            _index.append(self.X.columns.get_loc(i))
        if _index == []:
            raise NameError("No coincide ninguna de las features introducidas")
        return _index

    def get_important_features(self, display=True):

        self.model.get_feature_importance(prettified=True)
        _feature_importance_df = self.model.get_feature_importance(prettified=True)

        if display:
            plt.figure(figsize=(12, 6))
            sns.barplot(x="Importances", y="Feature Id", data=_feature_importance_df)
            plt.title('CatBoost features importance:')

        return _feature_importance_df

    def Visualizer_Models(self, directs=None, visu_model=True):

        directorios = []
        if len(directs) < 0:
            if visu_model:
                directorios.append(self.train_dir)
            else:
                raise NameError("No se ha seleccionado ningun directorio")
        else:
            if visu_model:
                directorios.append(self.train_dir)
            for i in directs:
                directorios.append(i)
        print(directorios)
        widget = MetricVisualizer(directorios)
        widget.start()

    def hyperopt_objective(self, params):

        _model = CatBoostClassifier(
            l2_leaf_reg=int(params['l2_leaf_reg']),
            learning_rate=params['learning_rate'],
            bagging_temperature=params["bagging_temperature"],
            iterations=500,
            eval_metric='AUC',
            random_seed=99,
            verbose=False,
            loss_function='Logloss'
            )
        _cv_data = catboost.cv(
            self.all_train_data,
            _model.get_params()
        )
        best_accuracy = np.max(_cv_data['test-AUC-mean'])

        return 1-best_accuracy

    def FineTune_hyperopt(self, X, y, mute=False):

        self.dataset(X, y)

        params_space = {
            'l2_leaf_reg': hyperopt.hp.qloguniform('l2_leaf_reg', 0, 2, 1),
            'learning_rate': hyperopt.hp.uniform('learning_rate', 1e-3, 5e-1),
            'bagging_temperature': hyperopt.hp.uniform("bagging_temperature", 0, 0.3)
            }
        trials = hyperopt.Trials()
        best = hyperopt.fmin(
            self.hyperopt_objective,
            space=params_space,
            algo=hyperopt.tpe.suggest,
            max_evals=2,
            trials=trials,
            rstate=RandomState(self.random_state)
        )
        if not mute:
            print("\nBest parameters:")
            print(best)
            print("\n")

        _parameters = self.params
        _parameters.update(best)

        _model = CatBoostClassifier(**_parameters)
        _cv_data = catboost.cv(self.all_train_data, _model.get_params())

        if not mute:
            print('\nPrecise validation accuracy score: {}'.format(np.max(_cv_data['test-Accuracy-mean'])))
        return best

    def FineTune_sklearn(self, X, y, mute=False, n_splits=10, n_iter=2):
        """
        https://www.kaggle.com/ksaaskil/pets-definitive-catboost-tuning
        """
        self.dataset(X, y)

        def build_search(modelo, param_distributions, cv=5, n_iter=10, verbose=1, random_state=99):
            """
            Builder function for RandomizedSearch.
            """
            QWS = make_scorer(cohen_kappa_score, weights='quadratic')
            return RandomizedSearchCV(modelo,
                                      param_distributions=param_distributions,
                                      cv=cv,
                                      return_train_score=True,
                                      refit='cohen_kappa_quadratic',
                                      n_iter=n_iter,
                                      n_jobs=None,
                                      scoring={
                                            'accuracy': make_scorer(accuracy_score),
                                            'cohen_kappa_quadratic': QWS
                                      },
                                      verbose=verbose,
                                      random_state=random_state)
        def pretty_cv_results(cv_results,
                              sort_by='rank_test_cohen_kappa_quadratic',
                              sort_ascending=True,
                              n_rows=30):
            """
            Return pretty Pandas dataframe from the `cv_results_` attribute of finished parameter search,
            ranking by test performance and only keeping the columns of interest.
            """
            df = pd.DataFrame(cv_results)
            cols_of_interest = [key for key in df.keys()
                                    if key.startswith('param_')
                                        or key.startswith("mean_train")
                                        or key.startswith("std_train")
                                        or key.startswith("mean_test")
                                        or key.startswith("std_test")
                                        or key.startswith('mean_fit_time')
                                        or key.startswith('rank')]
            return df.loc[:, cols_of_interest].sort_values(by=sort_by, ascending=sort_ascending).head(n_rows)

        def run_search(X_train, y_train, search, mute=False):
            search.fit(X_train, y_train)
            print('Best score is:', search.best_score_)
            return pretty_cv_results(search.cv_results_)

        param_distributions = {

            'iterations': [100, 200],
            'learning_rate': scipy.stats.uniform(0.01, 0.3),
            'max_depth': scipy.stats.randint(3, 10),
            'one_hot_max_size': [30],
            'l2_leaf_reg': scipy.stats.reciprocal(a=1e-2, b=1e1),
        }

        if mute:
            _verbose = 0
        else:
            _verbose = 1

        self.params.update({
                            'use_best_model': False
                            })
        _model = CatBoostClassifier(**self.params)

        catboost_search = build_search(_model,
                                        param_distributions=param_distributions,
                                        n_iter=n_iter,
                                        verbose=_verbose,
                                        cv=RepeatedStratifiedKFold(n_splits=n_splits, n_repeats=1, random_state=self.random_state))
        catboost_cv_results = run_search(self.X, self.y,
                                         search=catboost_search,
                                         mute=mute)
        best_estimator = catboost_search.best_estimator_
        if not mute:
            print(best_estimator.get_params())

        return catboost_cv_results, best_estimator

    def __getattr__(self, attr):
        """
        Pass all other method calls to self.model.
        """
        return getattr(self.model, attr)

"""
POR IMPLEMENTAR E INTEGRAR .....

#     def explainers_Shap(self):
#         
#         SHAP (SHapley Additive exPlanations) is a game theoretic approach
#         to explain the output of any machine learning model. It connects
#         optimal credit allocation with local explanations using the
#         classic Shapley values from game theory and their related extensions
#         
#         _cbc = CatBoostClassifier(**self.params)
        
#         _cbc.fit(self.train_data, # instead of X_train, y_train
#                       eval_set=self.eval_data, # instead of (X_valid, y_valid)
#                       use_best_model=True,
#                       plot=False)
        
#         _explainer = shap.TreeExplainer(_cbc) # insert your model
#         shap_values = _explainer.shap_values(self.train_data) # insert your train Pool object
        
        # Los primeros 100 valores
#         shap.initjs()
#         shap.force_plot(_explainer.expected_value, shap_values[:100,:], self.X_train.iloc[:100,:])
#         shap.summary_plot(shap_values, self.X_train)

"""



"""
Jaime Sendra Berenguer-2020.
MLearner Machine Learning Library Extensions
Author:Jaime Sendra Berenguer<www.linkedin.com/in/jaisenbe>
License: MIT
"""

from .load_data_images import load_image, plot_image, plot_image2

__all__ = ["load_image", "plot_image", "plot_image2"]
